# TODO - implement vLLM provider for faster inference

"""
This is a subclass (VLLM) for abstract class (BaseLLM) that implements an interface 
to interact with downloaded text generation models (compatible with HuggingFace models). 

A YAML configuration file is required to specify model parameters and other 
provider-specific settings. By default, the l2p library includes a configuration file 
located at 'l2p/llm/utils/llm.yaml'.

Users can also define their own custom models and parameters by extending the YAML 
configuration using the same format template.

There are two variations that users can use vLLM:
    1. Native vLLM
    2. OpenAI-Compatible Server (currently not supported by l2p)
"""

from typing_extensions import override
from .base import BaseLLM, load_yaml
from .utils.prompt_template import prompt_templates

class VLLM(BaseLLM):
    def __init__(
            self, 
            model, 
            model_path: str, # base directory of stored model
            config_path: str = "l2p/llm/utils/llm.yaml",
            provider: str = "huggingface",
            api_key: str | None = None, # only if model is affiliated w/ private repo
        ) -> None:

        try:
            from vllm import LLM, SamplingParams
        except ImportError:
            raise ImportError(
                "The 'vllm' library is required for VLLM but is not installed or "
                "failed to import properly. Install it using: `pip install vllm`."
            )
        
        self.api_key = api_key
        
        # load yaml configuration path
        self.provider = provider
        self._config = load_yaml(config_path)
            
        # retrieve model configurations
        model_config = self._config.get(self.provider, {}).get(model, {})
        self.model_engine = model_config.get("engine", model)
        self.model_path = model_path

        # set parameters for model
        self._set_parameters(model_config)

        # set model configurations
        self._set_configs(model_config)

        self.sampling_params = SamplingParams(
            temperature = self.temperature,
            top_p = self.top_p,
            stop = self.stop,
            max_tokens = self.max_new_tokens,
        )

        # recording logs
        self.in_tokens = 0
        self.out_tokens = 0
        self.query_log = []

        if self.context_length > 8192:
            self.llm = LLM(
                model = self.model_path,
                dtype = self.dtype,
                tensor_parallel_size = self.ngpu,
                gpu_memory_utilization = 0.9,
                max_num_batched_tokens = 8192,
                max_model_len = 8192
                )
        else:
            self.llm = LLM(
                model = self.model_path,
                dtype = self.dtype,
                tensor_parallel_size = self.ngpu,
                gpu_memory_utilization = 0.9,
                max_num_batched_tokens = self.context_length,
                )
        
        self.tokenizer = self.llm.get_tokenizer()
    

    def _set_parameters(self, model_config: dict) -> None:
        """Set parameters from the model configuration"""
        
        # default values for parameters if none exists
        defaults = {
            "context_length": 4096,
            "max_new_tokens": 2048,
            "temperature": 0.0,
            "top_p": 1.0,
            "stop": None,
            "do_sample": False
        }
        
        parameters = model_config.get("model_params", {})
        for key, default in defaults.items():
            setattr(self, key, parameters.get(key, default))


    def _set_configs(self, model_config: dict) -> None:
        """Set model hardware configuration."""

        # mapping from string to torch.dtype
        dtype_map = {
            "float32": self.torch.float32,
            "float16": self.torch.float16,
            "bfloat16": self.torch.bfloat16,
            "int8": self.torch.int8,
        }

        # extract inner config if it exists
        configs = model_config.get("model_config", {})

        # get and parse torch_dtype
        d_type = configs.get("dtype", "float32")
        if isinstance(d_type, str):
            if d_type in dtype_map:
                self.dtype = dtype_map[d_type]
            else:
                raise ValueError(f"Unsupported dtype string: '{d_type}'. Must be one of {list(dtype_map.keys())}.")
        elif isinstance(d_type, self.torch.dtype):
            self.dtype = d_type
        else:
            raise TypeError("dtype must be a string or torch.dtype instance.")
        
        ngpu = configs.get("ngpu", 1)
        if isinstance(ngpu, int):
            self.ngpu = ngpu
        else:
            raise TypeError("ngpu must be an integer.")
        

    def generate_prompt(self, system_message, prompt):
        """Generate prompt structure for specific LLM."""
        system_message = system_message or "You are a PDDL coding assistant. Provide concise, correct code only.\n"
        model_name = self.model_engine.lower()

        # try to find matching template key from the prompt_templates
        for key in prompt_templates:
            if key in model_name:
                template = prompt_templates[key]
                return template.format(system_prompt=system_message, prompt=prompt).strip()

        # default fallback if no template matched
        return prompt
    
    @override
    def query(
        self, 
        prompt: str,
        system_prompt: str = None,
        end_when_error: bool=False,
        max_retry: int=3,
        est_margin: int=200,
        ) -> str:
        """Generate a response from model based on the prompt."""
        
        if not isinstance(prompt, str) or not prompt.strip():
            raise ValueError("Prompt must be a non-empty string.")
        
        full_prompt = self.generate_prompt(system_prompt, prompt)
        assert full_prompt is not None

        input = self.tokenizer(full_prompt)
        requested_tokens = len(input["input_ids"])

        available_context = self.context_length - requested_tokens - est_margin
        max_new_tokens = min(self.max_new_tokens, max(0, available_context))

        print(
            f"Requesting {max_new_tokens} tokens "
            f"(estimated prompt: {requested_tokens} tokens, margin: {est_margin}, window: {self.context_length})"
        )

        # request response
        conn_success, n_retry = False, 0
        while not conn_success and n_retry < max_retry:
            try:
                # print token information
                print(f"[INFO] connecting to {self.model_engine} ({requested_tokens} tokens)...")
                if requested_tokens >= self.context_length:
                    print(
                        f"[WARNING] Prompt is {requested_tokens} tokens and exceeds context length "
                        f"({self.context_length}). It will be truncated."
                    )
                
                llm_output = self.llm.generate([full_prompt], self.sampling_params)
                llm_output = llm_output[0].outputs[0].text
                    
                conn_success = True

            except Exception as e:
                print(f"[ERROR] LLM error: {e}")
                if end_when_error:
                    break
                n_retry += 1
        
        if not conn_success:
            raise ConnectionError(
                f"Failed to generate response after {max_retry} retries."
            )
        
        # retrieve output tokens
        output_ids = self.tokenizer(llm_output)
        output_tokens = len(output_ids["input_ids"])

        # record token counts
        self.in_tokens += requested_tokens
        self.out_tokens += output_tokens

        self.query_log.append({
            "model": self.model_engine,
            "prompt_tokens": requested_tokens,
            "completion_tokens": output_tokens,
            "total_tokens": requested_tokens + output_tokens,
            "prompt": full_prompt,
            "output": llm_output,
        })
    
        return llm_output
    
    def get_tokens(self) -> tuple[int,int]:
        """Return input and output token counts."""
        return self.in_tokens, self.out_tokens
    
    def reset_tokens(self) -> None:
        """Reset token counts."""
        self.in_tokens = 0
        self.out_tokens = 0

    def get_query_log(self) -> list:
        """Retrieve query log."""
        return self.query_log
    
    def reset_query_log(self) -> None:
        """Reset query log."""
        self.query_log = []

    @override
    def valid_models(self) -> list[str]:
        """Returns a list of valid model engines."""
        try:
            return list(self._config.get(self.provider, {}).keys())
        except KeyError:
            return []
            


"""
This module contains a collection of helper functions that parses information from LLM output.
"""

import ast
import json
import os
import re
import sys

from collections import OrderedDict
from copy import deepcopy
from typing import Optional

from pddl import parse_domain, parse_problem
from pddl.formatter import domain_to_string, problem_to_string

from .pddl_format import remove_comments
from .pddl_types import Action, Function, Predicate


# ---- PDDL DOMAIN PARSERS ----


def parse_types(llm_output: str, heading: str = "TYPES") -> Optional[dict[str, str]]:
    """
    Safely extracts and evaluates a dictionary structure from a string (LLM response).

    Args:
        llm_output (str): raw string from the LLM expected to contain a flat dictionary
        heading (str): header string to extract text. Defaults to `TYPES`

    Returns:
        types_parsed (dict[str, str]) | None: parsed dictionary of :types if valid, else None.
    """
    try:

        types_head = parse_heading(llm_output, heading)
        if types_head.count("```") != 2:
            raise ValueError(
                "Could not find exactly one block in the types section enclosed by [```, ```] of the LLM output."
            )

        types_raw = combine_blocks(types_head)

        # Regex to extract the first dictionary-like structure
        dict_pattern = re.compile(r"{[^{}]*}", re.DOTALL)
        match = dict_pattern.search(types_raw)

        if not match:
            print("No dictionary found in the LLM response.")
            return None

        dict_str = match.group(0)
        types_parsed = ast.literal_eval(dict_str)

        # Validate it is a flat dictionary with string keys and values
        if isinstance(types_parsed, dict) and all(
            isinstance(k, str) and isinstance(v, str) for k, v in types_parsed.items()
        ):
            return types_parsed

        print("Parsed object is not a flat dictionary of string keys and values.")
        return None

    except Exception as e:
        print(f"Error parsing dictionary: {e}")
        return None


def parse_type_hierarchy(llm_output: str, heading: str = "TYPES") -> Optional[list[dict[str, str]]]:
    """
    Safely parses LLM response into a list of nested dictionaries representing the type hierarchy.

    Args:
        llm_output (str): raw LLM output expected to contain a Python list of dictionaries.
        heading (str): header string to extract text. Defaults to `TYPES`

    Returns:
        types_parsed (list[dict[str,str]]) | None: parsed type hierarchy if valid, else None.
    """
    try:

        types_head = parse_heading(llm_output, heading)
        if types_head.count("```") != 2:
            raise ValueError(
                "Could not find exactly one block in the types section enclosed by [```, ```] of the LLM output."
            )

        types_raw = combine_blocks(types_head)

        # catch if types is empty
        if not types_raw:
            return None

        types_parsed = ast.literal_eval(types_raw)

        # ensure it's a list of dicts with proper structure
        if not isinstance(types_parsed, list):
            return None

        def validate_type_node(node):
            if not isinstance(node, dict):
                return False
            if "children" not in node:
                return False
            if not isinstance(node["children"], list):
                return False
            for child in node["children"]:
                if not validate_type_node(child):
                    return False
            return True

        if all(validate_type_node(item) for item in types_parsed):
            return types_parsed

        return None

    except Exception as e:
        print(f"Failed to convert response to dict: {e}")
        return None


def parse_constants(llm_output: str, heading: str = "CONSTANTS") -> Optional[dict[str, str]]:
    """
    Safely extracts and evaluates a dictionary structure from a string (LLM response).

    Args:
        llm_output (str): Raw string from the LLM expected to contain a flat dictionary.
        heading (str): header string to extract text. Defaults to `CONSTANTS`

    Returns:
        constants_parsed (dict[str, str]) | None: parsed dictionary if valid, else None.
    """
    try:

        constant_head = parse_heading(llm_output, heading)
        if constant_head.count("```") != 2:
            raise ValueError(
                "Could not find exactly one block in the constants section enclosed by [```, ```] of the LLM output."
            )

        constants_raw = combine_blocks(constant_head)

        # regex to extract the first dictionary-like structure
        dict_pattern = re.compile(r"{[^{}]*}", re.DOTALL)
        match = dict_pattern.search(constants_raw)

        if not match:
            print("No dictionary found in the LLM response.")
            return None

        dict_str = match.group(0)
        constants_parsed = ast.literal_eval(dict_str)

        # validate it is a flat dictionary with string keys and values
        if isinstance(constants_parsed, dict) and all(
            isinstance(k, str) and isinstance(v, str)
            for k, v in constants_parsed.items()
        ):
            return constants_parsed

        print("Parsed object is not a flat dictionary of string keys and values.")
        return None

    except Exception as e:
        print(f"Error parsing dictionary: {e}")
        return None


def parse_predicates(all_predicates: list[Predicate]) -> list[Predicate]:
    """
    This function assumes the predicate definitions adhere to PDDL grammar.
    Assigns `params` to the predicate arguments properly. This should be run
    after retrieving a predicate list to ensure predicates are set correctly.
    """
    all_predicates = deepcopy(all_predicates)
    for i, pred in enumerate(all_predicates):
        if "params" in pred:
            continue
        pred_def = pred["raw"].split(": ")[0]
        pred_def = pred_def.strip(" ()`")  # drop any leading/strange formatting
        split_predicate = pred_def.split(" ")[1:]  # discard the predicate name
        split_predicate = [e for e in split_predicate if e != ""]

        pred["params"] = OrderedDict()
        for j, p in enumerate(split_predicate):
            if j % 3 == 0:
                assert "?" in p, f"invalid predicate definition: {pred_def}"
                assert (
                    split_predicate[j + 1] == "-"
                ), f"invalid predicate definition: {pred_def}"
                param_name, param_obj_type = p, split_predicate[j + 2]
                pred["params"][param_name] = param_obj_type
    return all_predicates


def parse_new_predicates(llm_output: str, heading: str = "New Predicates") -> list[Predicate]:
    """
    Parses new predicates from LLM into Python format (refer to example templates to see
    how these predicates should be formatted in LLM response).

    LLM output header should contain '### New Predicates' along with structured content.
    """
    new_predicates = list()

    predicates_head = parse_heading(llm_output, heading)
    if predicates_head.count("```") != 2:
        raise ValueError(
            "Could not find exactly one block in the new predicates section enclosed by [```, ```] of the LLM output."
        )
    predicate_output = combine_blocks(predicates_head)

    for p_line in predicate_output.split("\n"):
        p_line = p_line.strip()
        if not p_line or p_line.startswith("```"):
            continue  # skip empty lines and code block markers

        # skip lines that do not look like predicate definitions
        if not (p_line.startswith("-") or p_line.startswith("(")):
            if len(p_line) > 0:
                print(f'[WARNING] unable to parse the line: "{p_line}"')
            continue

        # extract predicate signature and description
        if ":" in p_line:
            pred_part, desc = p_line.split(":", 1)
            predicate_desc = desc.strip().strip("'\"")
        elif ";" in p_line:
            pred_part, desc = p_line.split(";", 1)
            predicate_desc = desc.strip().strip("'\"")
        else:
            pred_part = p_line
            predicate_desc = ""

        # clean the predicate signature
        pred_part = pred_part.strip("- ()").strip()

        # split into name and parameters
        parts = pred_part.split()
        if not parts:
            continue

        predicate_name = parts[0]
        param_parts = parts[1:]

        params = OrderedDict()
        current_param = None

        i = 0
        while i < len(param_parts):
            part = param_parts[i]
            if part.startswith("?"):
                # found a parameter
                current_param = part
                params[current_param] = ""  # Default to untyped
                i += 1
            elif part == "-":
                # found type indicator
                if current_param is None:
                    print(
                        f"[WARNING] Found type indicator without parameter in: {p_line}"
                    )
                    i += 1
                    continue
                if i + 1 < len(param_parts):
                    params[current_param] = param_parts[i + 1]
                    i += 2
                else:
                    print(f"[WARNING] Missing type after '-' in: {p_line}")
                    i += 1
            else:
                # might be a typeless parameter (accept it with warning)
                print(
                    f"[WARNING] Assuming '{part}' is an untyped parameter in: {p_line}"
                )
                current_param = f"?{part.lstrip('?')}"
                params[current_param] = ""
                i += 1

        # generate clean PDDL representation
        clean_params = []
        for param, type_ in params.items():
            if type_ and type_ != "object":
                clean_params.append(f"{param} - {type_}")
            else:
                clean_params.append(param)
        clean = f"({predicate_name} {' '.join(clean_params)})"

        new_predicates.append(
            {
                "name": predicate_name,
                "desc": predicate_desc,
                "raw": p_line,
                "params": params,
                "clean": clean,
            }
        )

    return new_predicates


def parse_functions(llm_output: str, heading: str = "FUNCTIONS") -> list[Function]:
    """
    Parses function from LLM into Python format (refer to example templates to see
    how these functions should be formatted in LLM response).

    LLM output header should contain '### FUNCTIONS' along with structured content.
    """
    functions = list()
    functions_head = parse_heading(llm_output, heading)
    if functions_head.count("```") != 2:
        raise ValueError(
            "Could not find exactly one block in the functions section enclosed by [```, ```] of the LLM output."
        )
    function_output = combine_blocks(functions_head)

    for f_line in function_output.split("\n"):
        f_line = f_line.strip()
        if not f_line or f_line.startswith("```"):
            continue  # skip empty lines and code block markers

        # skip lines that do not look like function definitions
        if not (f_line.startswith("-") or f_line.startswith("(")):
            if len(f_line) > 0:
                print(f'[WARNING] unable to parse the line: "{f_line}"')
            continue

        # extract function signature and description
        if ":" in f_line:
            func_part, desc = f_line.split(":", 1)
            function_desc = desc.strip().strip("'\"")
        elif ";" in f_line:
            func_part, desc = f_line.split(";", 1)
            function_desc = desc.strip().strip("'\"")
        else:
            func_part = f_line
            function_desc = ""

        # clean the function signature
        is_numeric = False
        if re.search(r"- number\s*$", func_part):
            is_numeric = True
            func_part = re.sub(r"\s+- number\s*$", "", func_part)
        func_part = func_part.strip("- ()").strip()

        # split into name and parameters
        parts = func_part.split()
        if not parts:
            continue

        function_name = parts[0]
        param_parts = parts[1:]

        params = OrderedDict()
        current_param = None

        i = 0
        while i < len(param_parts):
            part = param_parts[i]
            if part.startswith("?"):
                # found a parameter
                current_param = part
                params[current_param] = ""  # default to untyped
                i += 1
            elif part == "-":
                # found type indicator
                if current_param is None:
                    print(
                        f"[WARNING] Found type indicator without parameter in: {f_line}"
                    )
                    i += 1
                    continue
                if i + 1 < len(param_parts):
                    params[current_param] = param_parts[i + 1]
                    i += 2
                else:
                    print(f"[WARNING] Missing type after '-' in: {f_line}")
                    i += 1
            else:
                # might be a typeless parameter (accept it with warning)
                print(
                    f"[WARNING] Assuming '{part}' is an untyped parameter in: {f_line}"
                )
                current_param = f"?{part.lstrip('?')}"
                params[current_param] = ""
                i += 1

        # generate clean PDDL representation
        clean_params = []
        for param, type_ in params.items():
            if type_:
                clean_params.append(f"{param} - {type_}")
            else:
                clean_params.append(param)
        
        clean = f"({function_name} {' '.join(clean_params)})"

        if is_numeric:
            clean += " - number"

        functions.append(
            {
                "name": function_name,
                "desc": function_desc,
                "raw": f_line,
                "params": params,
                "clean": clean,
            }
        )

    return functions


def parse_action(
        llm_output: str, 
        action_name: str, 
        param_head: str = "Action Parameters", 
        precon_head: str = "Action Preconditions",
        effect_head: str = "Action Effects") -> Action:
    """
    Parse an action from a given LLM output.

    Args:
        llm_output (str): raw LLM output
        action_name (str): the name of the action
        param_head (str): parameter header for extraction
        precon_head (str): precondition header for extraction
        effect_head (str): effect header for extraction

    Returns:
        Action: the parsed PDDL action
    """
    parameters, _ = parse_params(llm_output, param_head)
    preconditions = parse_preconditions(llm_output, precon_head)
    effects = parse_effects(llm_output, effect_head)

    return {
        "name": action_name,
        "params": parameters,
        "preconditions": preconditions,
        "effects": effects,
    }


def parse_params(llm_output: str, heading: str = "Action Parameters") -> tuple[OrderedDict, list]:
    """
    Parses parameters from LLM into Python format (refer to example templates to see
    how these parameters should be formatted in LLM response).

    LLM output header should contain 'Parameters' along with structured content.
    """
    params_info = OrderedDict()
    params_head = parse_heading(llm_output, heading)
    if params_head.count("```") != 2:
        raise ValueError(
            "Could not find exactly one block in the parameters section enclosed by [```, ```] of the LLM output."
        )
    params_str = combine_blocks(params_head)
    params_raw = []

    for line in params_str.split("\n"):
        line = line.strip()
        if not line:  # skip empty lines
            continue

        if line.startswith("-"):
            line = line[1:].strip()  # remove the dash and clean up

        if not line.startswith("?"):
            print(f"[WARNING] Invalid parameter line. Must start with '?': '{line}'")
            continue

        try:
            params_raw.append(line)
            # split into param name and type
            parts = line.split(":")

            left_side = parts[0].strip()
            param_parts = [p.strip() for p in left_side.split(" - ")]
            param_name = param_parts[0].strip(" `").strip("-").strip()

            if len(param_parts) == 2 and param_parts[1]:
                param_type = param_parts[1].strip(" `")
            else:
                param_type = "object"  # no type provided

            params_info[param_name] = param_type

        except Exception as e:
            print(f"[WARNING] Failed to parse parameter line: '{line}' - {str(e)}")
            continue

    return params_info, params_raw


def parse_preconditions(llm_output: str, heading: str = "Action Preconditions") -> str:
    """Parses precondition string from LLM output"""
    try:
        preconds_head = parse_heading(llm_output, heading)
        if preconds_head.count("```") != 2:
            raise ValueError(
                "Could not find exactly one block in the preconditions section enclosed by [```, ```] of the LLM output."
            )
        preconditions = combine_blocks(preconds_head)

        return preconditions
    except:
        raise Exception(
            "Could not find the 'Preconditions' section in the output. Provide the entire response, including all headings even if some are unchanged."
        )


def parse_effects(llm_output: str, heading: str = "Action Effects") -> str:
    """Parses effect string from LLM output"""
    try:
        effects_head = parse_heading(llm_output, heading)
        if effects_head.count("```") != 2:
            raise ValueError(
                "Could not find exactly one block in the effects section enclosed by [```, ```] of the LLM output."
            )
        effects = combine_blocks(effects_head)

        return effects
    except:
        raise Exception(
            "Could not find the 'Preconditions' section in the output. Provide the entire response, including all headings even if some are unchanged."
        )


def prune_types(
    types: dict[str, str] | list[dict[str, str]],
    predicates: list[Predicate],
    actions: list[Action],
) -> dict[str, str]:
    """
    Prune types that are not used in any predicate or action.

    Parameters:
        types (dict or list): Either a flat dict of {type: description} or a nested list of type hierarchies.
        predicates (list[Predicate]): A list of predicates.
        actions (list[Action]): A list of actions.

    Returns:
        dict[str, str]: A dictionary of used types.
    """
    used_types = {}
    all_type_names = {}

    def collect_all_types(obj):
        """Recursively extract all type names and their descriptions from the input."""

        # if setup is normal types dictionary
        if isinstance(obj, dict):
            for type_name, desc in obj.items():
                all_type_names[type_name] = desc

        # if set up is type hierarchy list dictionary
        elif isinstance(obj, list):
            for node in obj:
                type_name = next((k for k in node if k != "children"), None)
                if type_name:
                    all_type_names[type_name] = node[type_name]
                    collect_all_types(node.get("children", []))

    # flatten all types into all_type_names
    collect_all_types(types)

    # identify used types
    for type_name in all_type_names:
        base_name = type_name.split(" ")[0]
        found = False

        for pred in predicates:
            if base_name in pred["params"].values():
                used_types[type_name] = all_type_names[type_name]
                found = True
                break
        if found:
            continue

        for action in actions:
            if base_name in action["params"].values():
                used_types[type_name] = all_type_names[type_name]
                break
            if base_name in action["preconditions"] or base_name in action["effects"]:
                used_types[type_name] = all_type_names[type_name]
                break

    return used_types


def prune_predicates(
    predicates: list[Predicate], actions: list[Action]
) -> list[Predicate]:
    """
    Remove predicates that are not used in any action.

    Args:
        predicates (list[Predicate]): a list of predicates
        actions (list[Action]): a list of actions

    Returns:
        list[Predicate]: the pruned list of predicates
    """
    used_predicates = []
    seen_predicate_names = set()

    for pred in predicates:
        for action in actions:
            # Add a space or a ")" to avoid partial matches
            names = [f"{pred['name']} ", f"{pred['name']})"]
            for name in names:
                if name in action["preconditions"] or name in action["effects"]:
                    if pred["name"] not in seen_predicate_names:
                        used_predicates.append(pred)
                        seen_predicate_names.add(pred["name"])
                    break

    return used_predicates


# ---- PDDL PROBLEM PARSERS ----


def parse_objects(llm_output: str, heading: str = "OBJECTS") -> dict[str, str]:
    """
    Extract objects from LLM response and returns dictionary string pairs object(name, type)

    Args:
        llm_output (str): raw LLM output
        types (dict[str,str]): WILL BE USED FOR CHECK ERROR RAISES
        predicates (list[Predicate]): WILL BE USED FOR CHECK ERROR RAISES

    Returns:
        objects_parsed (dict[str,str]): PDDL task objects
    """

    objects_head = parse_heading(llm_output, heading)
    if objects_head.count("```") != 2:
        raise ValueError(
            "Could not find exactly one block in the problem objects section enclosed by [```, ```] of the LLM output."
        )
    objects_raw = combine_blocks(objects_head)
    objects_clean = remove_comments(objects_raw)

    objects_parsed = {
        obj.split(" - ")[0].strip(" `"): obj.split(" - ")[1].strip(" `")
        for obj in objects_clean.split("\n")
        if obj.strip()
    }

    return objects_parsed


def parse_initial(llm_output: str, heading: str = "INITIAL") -> list[dict[str, str]]:
    """
    Extracts state (PDDL-init) from LLM response and returns it as a list of dict strings

    Parameters:
        llm_output (str): raw LLM output

    Returns:
        states (list[dict[str,str]]): list of initial states in dictionaries
    """
    initial_head = parse_heading(llm_output, heading)
    if initial_head.count("```") != 2:
        raise ValueError(
            "Could not find exactly one block in the initial section enclosed by [```, ```] of the LLM output. The initial states have to be specified in a single block Likely this is caused by a too long response and limited context length. If so, try to shorten the message and exclude objects which aren't needed for the task."
        )

    initial_raw = combine_blocks(initial_head)
    initial_clean = remove_comments(initial_raw)
    initial_parsed = parse_pddl(f"({initial_clean})")

    return parse_task_states(initial_parsed)


def parse_goal(llm_output: str, heading: str = "GOAL") -> list[dict[str, str]]:
    """
    Extracts goal (PDDL-goal) from LLM response and returns it as a string

    Parameters:
        llm_output (str): raw LLM output

    Returns:
        states (list[dict[str,str]]): list of goal states in dictionaries
    """
    goal_head = parse_heading(llm_output, heading)
    if goal_head.count("```") != 2:
        raise ValueError(
            "Could not find exactly one block in the goal section enclosed by [```, ```] of the LLM output. The goal states have to be specified in a single block Likely this is caused by a too long response and limited context length. If so, try to shorten the message and exclude objects which aren't needed for the task."
        )

    goal_raw = combine_blocks(goal_head)
    goal_clean = remove_comments(goal_raw)
    goal_parsed = parse_pddl(f"({goal_clean})")

    return parse_task_states(goal_parsed)


def parse_task_states(parsed_states: list) -> list[dict]:
    states = []
    for line in parsed_states:

        if isinstance(line, str):
            line = [line]

        # only parse lines
        if isinstance(line, list):
            name = line[0].split(" ")[0]

            if name == "and":
                continue
            # if comparsion operator
            if name in ["=", ">", "<", ">=", "<="]:
                op = name
                name = line[1][0].split(" ")[0]  # retrieve function name
                params = line[1][0].split(" ")[1:]
                value = line[2]
                states.append(
                    {"func_name": name, "params": params, "value": value, "op": op}
                )
                continue
            if name == "not":
                name = line[1][0].split(" ")[0]  # retrieve function name
                params = line[1][0].split(" ")[1:]
                neg = True
            else:
                neg = False
                params = line[0].split(" ")[1:]
            states.append({"pred_name": name, "params": params, "neg": neg})

    return states


# ---- SUPPLEMENTARY PARSING FUNCTIONS ----


def parse_heading(llm_output: str, heading: str) -> str:
    """Extract the text between the heading and the next second level heading in the LLM output."""
    if heading not in llm_output:
        raise ValueError(
            f"Could not find heading {heading} in the LLM output:\n{llm_output}\n. Likely this is caused by a too long response and limited context length. If so, try to shorten the message and exclude objects which aren't needed for the task"
        )
    heading_str = (
        llm_output.split(heading)[1].split("\n### ")[0].strip()
    )  # Get the text between the heading and the next heading
    return heading_str


def parse_pddl(pddl_str: str) -> list:
    """
    Simplified PDDL parser that converts the string into a nested list structure.
    """

    # tokenize string
    tokens = re.sub(r"([()])", r" \1 ", pddl_str).split()
    stack = []
    current = []

    # iterate each token, if it is a bracket, we append to new list
    for token in tokens:
        if token == "(":
            stack.append(current)
            current = []
        elif token == ")":
            if stack:
                parent = stack.pop()
                parent.append(current)
                current = parent
        else:
            current.append(token)

    if len(current) != 1:
        raise ValueError("Malformed PDDL expression")

    nested_pddl = concatenate_strings(current[0])
    return nested_pddl


def combine_blocks(heading_str: str):
    """Combine the inside of blocks from the heading string into a single string."""

    possible_blocks = heading_str.split("```")
    blocks = [
        possible_blocks[i] for i in range(1, len(possible_blocks), 2)
    ]  # obtain string between ```

    combined = "\n".join(blocks)

    return combined.replace(
        "\n\n", "\n"
    ).strip()  # remove leading/trailing whitespace and internal empty lines


def concatenate_strings(nested_list):
    """Helper function that concatenates strings within a list together."""
    if not isinstance(nested_list, list):
        return nested_list

    new_list = []
    i = 0
    while i < len(nested_list):
        current = nested_list[i]
        if isinstance(current, str):
            # start collecting consecutive strings
            concatenated = current
            j = i + 1
            while j < len(nested_list) and isinstance(nested_list[j], str):
                concatenated += " " + nested_list[j]
                j += 1
            new_list.append(concatenated)
            i = j
        else:
            # recurse for sublists
            new_list.append(concatenate_strings(current))
            i += 1
    return new_list

    # ---- SUPPLEMENTARY FUNCTIONS ----


# ---- HELPER FUNCTIONS ----


def load_file(file_path: str):
    """Helper function that loads a single file into a string"""
    _, ext = os.path.splitext(file_path)
    with open(file_path, "r") as file:
        if ext == ".json":
            return json.load(file)
        else:
            return file.read().strip()


def load_files(folder_path: str):
    """Helper function that loads multiple files into a string list"""
    file_contents = []
    for filename in sorted(os.listdir(folder_path)):
        file_path = os.path.join(folder_path, filename)
        if os.path.isfile(file_path):
            with open(file_path, "r") as file:
                file_contents.append(file.read())
    return file_contents


def check_parse_domain(file_path: str):
    """Run PDDL library to check if file is syntactically correct"""
    try:
        domain = parse_domain(file_path)
        pddl_domain = domain_to_string(domain)
        return pddl_domain
    except Exception as e:
        print("------------------")
        print(f"Error parsing domain: {e}", file=sys.stderr)
        print("------------------")
        sys.exit(1)


def check_parse_problem(file_path: str):
    """Run PDDL library to check if file is syntactically correct"""
    try:
        problem = parse_problem(file_path)
        pddl_problem = problem_to_string(problem)
        return pddl_problem
    except Exception as e:
        print("------------------")
        print(f"Error parsing domain: {e}", file=sys.stderr)
        print("------------------")
        sys.exit(1)



import unittest, textwrap
from l2p.task_builder import TaskBuilder
from l2p.utils import *
from l2p.utils.pddl_validator import SyntaxValidator
from l2p.utils.pddl_types import Predicate
from .mock_llm import MockLLM


class TestTaskBuilder(unittest.TestCase):
    def setUp(self):
        self.task_builder = TaskBuilder()
        self.syntax_validator = SyntaxValidator()
        self.mock_llm = MockLLM()

    def normalize(self, string):
        return "\n".join(
            line.strip() for line in textwrap.dedent(string).strip().splitlines()
        )

    def test_formalize_objects(self):

        self.syntax_validator.headers = ["OBJECTS"]
        self.syntax_validator.error_types = [
            "validate_header",
            "validate_duplicate_headers",
            "validate_unsupported_keywords",
            "validate_task_objects",
        ]
        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT
            ### OBJECTS
            ```
            object1 - type_1
            object2 - type_2
            object3 - type_3
            ```
            ## DISTRACTION TEXT
            """
        )

        types = {"type_1": "", "type_2": "", "type_3": ""}

        objects, _, validation_info = self.task_builder.formalize_objects(
            model=self.mock_llm,
            problem_desc="",
            prompt_template="",
            types=types,
            syntax_validator=self.syntax_validator,
        )

        exp_objects = {"object1": "type_1", "object2": "type_2", "object3": "type_3"}

        self.assertEqual(validation_info[0], True)
        self.assertEqual(exp_objects, objects)

    def test_formalize_initial_state(self):

        self.syntax_validator.headers = ["INITIAL"]
        self.syntax_validator.error_types = [
            "validate_header",
            "validate_duplicate_headers",
            "validate_unsupported_keywords",
            "validate_task_states",
        ]
        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT
            ### INITIAL
            ```
            (predicate_1 object1 object2) ; comment for initial state predicate 1
            (predicate_2 object3 object1) ; comment for initial state predicate 2
            (predicate_3 object1) ; comment for initial state predicate 3
            ```
            ## DISTRACTION TEXT
            """
        )

        types = {"type_1": "", "type_2": "", "type_3": ""}
        objects = {"object1": "type_1", "object2": "type_2", "object3": "type_3"}

        predicates = [
            Predicate(
                {
                    "name": "predicate_1",
                    "desc": "",
                    "raw": "(predicate_1 ?t1 ?t2): description",
                    "params": OrderedDict([("?t1", "type_1"), ("?t2", "type_2")]),
                    "clean": "(predicate_1 ?t1 ?t2",
                }
            ),
            Predicate(
                {
                    "name": "predicate_2",
                    "desc": "",
                    "raw": "(predicate_2 ?t1 ?t2): description",
                    "params": OrderedDict([("?t1", "type_3"), ("?t2", "type_1")]),
                    "clean": "(predicate_2 ?t1 ?t2",
                }
            ),
            Predicate(
                {
                    "name": "predicate_3",
                    "desc": "",
                    "raw": "(predicate_3 ?t1): description",
                    "params": OrderedDict([("?t1", "type_1")]),
                    "clean": "(predicate_3 ?t1",
                }
            ),
        ]

        initial, _, validation_info = self.task_builder.formalize_initial_state(
            model=self.mock_llm,
            problem_desc="",
            prompt_template="",
            types=types,
            predicates=predicates,
            objects=objects,
            syntax_validator=self.syntax_validator,
        )

        exp_initial = [
            {
                "pred_name": "predicate_1",
                "params": ["object1", "object2"],
                "neg": False,
            },
            {
                "pred_name": "predicate_2",
                "params": ["object3", "object1"],
                "neg": False,
            },
            {"pred_name": "predicate_3", "params": ["object1"], "neg": False},
        ]

        self.assertEqual(validation_info[0], True)
        self.assertEqual(exp_initial, initial)

    def test_formalize_task(self):
        self.syntax_validator.headers = ["INITIAL"]
        self.syntax_validator.error_types = [
            "validate_header",
            "validate_duplicate_headers",
            "validate_unsupported_keywords",
            "validate_task_states",
            "validate_task_objects",
            "validate_task_states",
        ]
        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT
            ### OBJECTS
            ```
            object1 - type_1
            object2 - type_2
            object3 - type_3
            ```

            ## DISTRACTION TEXT
            ### INITIAL
            ```
            (predicate_1 object1 object2) ; comment for initial state predicate 1
            (predicate_2 object3 object1) ; comment for initial state predicate 2
            ```

            ## DISTRACTION TEXT
            ### GOAL
            ```
            (predicate_3 object1) ; comment for initial state predicate 3
            (not (predicate_2 object3 object1))
            ```
            ## DISTRACTION TEXT
            """
        )

        types = {"type_1": "", "type_2": "", "type_3": ""}

        predicates = [
            Predicate(
                {
                    "name": "predicate_1",
                    "desc": "",
                    "raw": "(predicate_1 ?t1 ?t2): description",
                    "params": OrderedDict([("?t1", "type_1"), ("?t2", "type_2")]),
                    "clean": "(predicate_1 ?t1 ?t2",
                }
            ),
            Predicate(
                {
                    "name": "predicate_2",
                    "desc": "",
                    "raw": "(predicate_2 ?t1 ?t2): description",
                    "params": OrderedDict([("?t1", "type_3"), ("?t2", "type_1")]),
                    "clean": "(predicate_2 ?t1 ?t2",
                }
            ),
            Predicate(
                {
                    "name": "predicate_3",
                    "desc": "",
                    "raw": "(predicate_3 ?t1): description",
                    "params": OrderedDict([("?t1", "type_1")]),
                    "clean": "(predicate_3 ?t1",
                }
            ),
        ]

        objects, initial, goal, _, validation_info = self.task_builder.formalize_task(
            model=self.mock_llm,
            problem_desc="",
            prompt_template="",
            types=types,
            predicates=predicates,
            syntax_validator=self.syntax_validator,
        )

        exp_objects = {"object1": "type_1", "object2": "type_2", "object3": "type_3"}

        exp_initial = [
            {
                "pred_name": "predicate_1",
                "params": ["object1", "object2"],
                "neg": False,
            },
            {
                "pred_name": "predicate_2",
                "params": ["object3", "object1"],
                "neg": False,
            },
        ]

        exp_goal = [
            {"pred_name": "predicate_3", "params": ["object1"], "neg": False},
            {"pred_name": "predicate_2", "params": ["object3", "object1"], "neg": True},
        ]

        self.assertEqual(validation_info[0], True)
        self.assertEqual(exp_objects, objects)
        self.assertEqual(exp_initial, initial)
        self.assertEqual(exp_goal, goal)

    def test_generate_task(self):

        objects = {"object1": "type_1", "object2": "type_2", "object3": "type_3"}
        initial = [
            {
                "pred_name": "predicate_1",
                "params": ["object1", "object2"],
                "neg": False,
            },
            {
                "pred_name": "predicate_2",
                "params": ["object3", "object1"],
                "neg": False,
            },
        ]
        goal = [{"pred_name": "predicate_3", "params": ["object1"], "neg": False}]

        expected_output = textwrap.dedent(
            """
            (define
                (problem test_problem)
                (:domain test_domain)

                (:objects 
                    object1 - type_1
                    object2 - type_2
                    object3 - type_3
                )

                (:init
                    (predicate_1 object1 object2)
                    (predicate_2 object3 object1)
                )

                (:goal
                    (and 
                        (predicate_3 object1) 
                    )
                )
            )
            """
        )

        result = self.task_builder.generate_task(
            domain_name="test_domain",
            problem_name="test_problem",
            objects=objects,
            initial=initial,
            goal=goal,
        )

        self.assertEqual(self.normalize(result), self.normalize(expected_output))


if __name__ == "__main__":
    unittest.main()



import unittest, textwrap
from collections import OrderedDict
from l2p import *
from .mock_llm import MockLLM


class TestDomainBuilder(unittest.TestCase):
    def setUp(self):
        self.domain_builder = DomainBuilder()
        self.syntax_validator = SyntaxValidator()
        self.mock_llm = MockLLM()

    def normalize(self, string):
        return "\n".join(
            line.strip() for line in textwrap.dedent(string).strip().splitlines()
        )

    def test_formalize_types(self):

        self.syntax_validator.error_types = ["validate_format_types"]
        self.mock_llm.output = textwrap.dedent(
            """
            ### TYPES
            ```
            {
                "arm": "arm for a robot",
                "block": "block that can be stacked and unstacked",
                "table": "table that blocks sits on",
            }
            ```
            """
        )

        expected_types = {
            "arm": "arm for a robot",
            "block": "block that can be stacked and unstacked",
            "table": "table that blocks sits on",
        }

        types, _, validation_info = self.domain_builder.formalize_types(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            syntax_validator=self.syntax_validator,
        )

        self.assertEqual(expected_types, types)
        self.assertEqual(validation_info[0], True)

    def test_formalize_type_hierarchy(self):
        self.syntax_validator.error_types = [
            "validate_format_types",
            "validate_cyclic_types",
        ]
        self.mock_llm.output = textwrap.dedent(
            """
            ### TYPES
            ```
            [
                {
                    "arm": "'arm for a robot'",
                    "children": []
                },
                {
                    "block": "block that can be stacked and unstacked",
                    "children": [
                        {
                            "heavy_block": "a heavy block that cannot be picked up",
                            "children": []
                        },
                        {
                            "light_block": "a light block that can be picked up",
                            "children": []
                        }
                    ]
                },
                {
                    "table": "table that blocks sits on",
                    "children": []
                }
            ]
            ```
            """
        )

        expected_types = [
            {"arm": "'arm for a robot'", "children": []},
            {
                "block": "block that can be stacked and unstacked",
                "children": [
                    {
                        "heavy_block": "a heavy block that cannot be picked up",
                        "children": [],
                    },
                    {
                        "light_block": "a light block that can be picked up",
                        "children": [],
                    },
                ],
            },
            {"table": "table that blocks sits on", "children": []},
        ]

        types, _, validation_info = self.domain_builder.formalize_type_hierarchy(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            syntax_validator=self.syntax_validator,
        )

        self.assertEqual(expected_types, types)
        self.assertEqual(validation_info[0], True)

    def test_formalize_functions(self):

        self.syntax_validator.error_types = ["validate_format_functions"]

        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT
            ### FUNCTIONS
            ```
            (battery-level ?a - arm): battery level of robot arm
            (weight ?b - block): weight of a block
            ```
            ## DISTRACTION TEXT
            """
        )

        types = {
            "arm": "arm for a robot",
            "block": "block that can be stacked and unstacked",
            "table": "table that blocks sits on",
        }

        functions, _, validation_info = self.domain_builder.formalize_functions(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            types=types,
            syntax_validator=self.syntax_validator,
        )

        exp_functions = [
            {
                "name": "battery-level",
                "desc": "battery level of robot arm",
                "raw": "(battery-level ?a - arm): battery level of robot arm",
                "params": OrderedDict([("?a", "arm")]),
                "clean": "(battery-level ?a - arm)",
            },
            {
                "name": "weight",
                "desc": "weight of a block",
                "raw": "(weight ?b - block): weight of a block",
                "params": OrderedDict([("?b", "block")]),
                "clean": "(weight ?b - block)",
            },
        ]

        self.assertEqual(validation_info[0], True)
        self.assertEqual(exp_functions, functions)

    def test_formalize_pddl_action(self):

        self.syntax_validator.unsupported_keywords = []

        self.syntax_validator.error_types = [
            "validate_header",
            "validate_duplicate_headers",
            "validate_unsupported_keywords",
            "validate_params",
            "validate_duplicate_predicates",
            "validate_types_predicates",
            "validate_format_predicates",
            "validate_usage_action",
        ]
        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT
            ### Action Parameters
            ```
            - ?b1 - block: The block being stacked on top
            - ?b2 - block: The block being stacked upon
            - ?a - arm: The arm performing the stacking action
            ```

            ## DISTRACTION TEXT

            ### Action Preconditions
            ```
            (and
                (holding ?a ?b1) ; The arm is holding the top block
                (clear ?b2) ; The bottom block is clear
                (= (weight ?b1) (weight ?b2))
                (forall (?block1 ?block2 - block ?a1 - arm)
                    (and
                    (clear ?block1)
                    (= (battery-level ?a) 100)
                    (>= (weight ?block1) 10)
                    (<= (weight ?b1) -100)
                    (< (weight ?block2) 20)
                    (> (battery-level ?a1) 10)
                    )
                )
            )
            ```

            ## DISTRACTION TEXT

            ### Action Effects
            ```
            (and
                (not (holding ?a ?b1)) ; The arm is no longer holding the top block
                (on ?b1 ?b2) ; The top block is now on the bottom block
                (not (clear ?b2)) ; The bottom block is no longer clear
                (when (clear ?b1)
                    (and
                    (clear ?b2)
                    (clear ?b1)
                    )
                )
                (increase (battery-level ?a) 200)
                (increase (battery-level ?a) (* (weight ?b1) 100))
                (decrease (battery-level ?a) (* 0.05 (weight ?b2)))
                (assign (battery-level ?a) (* (weight ?b1) 2))
                (scale-up (battery-level ?a) 2)
                (scale-down (battery-level ?a) (weight ?b1))
            )
            ```

            ## DISTRACTION TEXT

            ### New Predicates
            ```
            - (on ?b1 - block ?b2 - block): true if the block ?b1 is on top of the block ?b2
            - (holding ?a - arm ?b - block): true if arm is holding a block
            - (clear ?b - block): true if a block does not have anything on top of it
            ```

            ## DISTRACTION TEXT
            """
        )

        exp_predicates = [
            Predicate(
                {
                    "name": "on",
                    "desc": "true if the block ?b1 is on top of the block ?b2",
                    "raw": "- (on ?b1 - block ?b2 - block): true if the block ?b1 is on top of the block ?b2",
                    "params": OrderedDict([("?b1", "block"), ("?b2", "block")]),
                    "clean": "(on ?b1 - block ?b2 - block)",
                }
            ),
            Predicate(
                {
                    "name": "holding",
                    "desc": "true if arm is holding a block",
                    "raw": "- (holding ?a - arm ?b - block): true if arm is holding a block",
                    "params": OrderedDict([("?a", "arm"), ("?b", "block")]),
                    "clean": "(holding ?a - arm ?b - block)",
                }
            ),
            Predicate(
                {
                    "name": "clear",
                    "desc": "true if a block does not have anything on top of it",
                    "raw": "- (clear ?b - block): true if a block does not have anything on top of it",
                    "params": OrderedDict([("?b", "block")]),
                    "clean": "(clear ?b - block)",
                }
            ),
        ]

        exp_action = {
            "name": "stack",
            "params": OrderedDict([("?b1", "block"), ("?b2", "block"), ("?a", "arm")]),
            "preconditions": "(and\n    (holding ?a ?b1) ; The arm is holding the top block\n    (clear ?b2) ; The bottom block is clear\n    (= (weight ?b1) (weight ?b2))\n    (forall (?block1 ?block2 - block ?a1 - arm)\n        (and\n        (clear ?block1)\n        (= (battery-level ?a) 100)\n        (>= (weight ?block1) 10)\n        (<= (weight ?b1) -100)\n        (< (weight ?block2) 20)\n        (> (battery-level ?a1) 10)\n        )\n    )\n)",
            "effects": "(and\n    (not (holding ?a ?b1)) ; The arm is no longer holding the top block\n    (on ?b1 ?b2) ; The top block is now on the bottom block\n    (not (clear ?b2)) ; The bottom block is no longer clear\n    (when (clear ?b1)\n        (and\n        (clear ?b2)\n        (clear ?b1)\n        )\n    )\n    (increase (battery-level ?a) 200)\n    (increase (battery-level ?a) (* (weight ?b1) 100))\n    (decrease (battery-level ?a) (* 0.05 (weight ?b2)))\n    (assign (battery-level ?a) (* (weight ?b1) 2))\n    (scale-up (battery-level ?a) 2)\n    (scale-down (battery-level ?a) (weight ?b1))\n)",
        }

        types = {
            "arm": "arm for a robot",
            "block": "block that can be stacked and unstacked",
            "table": "table that blocks sits on",
        }

        functions = [
            {
                "name": "battery-level",
                "desc": "battery level of robot arm",
                "raw": "(battery-level ?a - arm): battery level of robot arm",
                "params": OrderedDict([("?a", "arm")]),
                "clean": "(battery-level ?a - arm)",
            },
            {
                "name": "weight",
                "desc": "weight of a block",
                "raw": "(weight ?b - block): weight of a block",
                "params": OrderedDict([("?b", "block")]),
                "clean": "(weight ?b - block)",
            },
        ]

        action, new_predicates, _, validation_info = (
            self.domain_builder.formalize_pddl_action(
                model=self.mock_llm,
                domain_desc="",
                prompt_template="",
                types=types,
                functions=functions,
                syntax_validator=self.syntax_validator,
                action_name="stack",
                extract_new_preds=True,
            )
        )

        self.assertEqual(exp_action, action)
        self.assertEqual(exp_predicates, new_predicates)
        self.assertEqual(validation_info[0], True)

    def test_formalize_parameters(self):

        self.syntax_validator.headers = ["Action Parameters"]
        self.syntax_validator.error_types = [
            "validate_header",
            "validate_duplicate_headers",
            "validate_unsupported_keywords",
            "validate_params",
        ]

        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT

            ### Action Parameters
            ```
            - ?b1 - block: the block that is being stacked on top
            - ?b2 - block: the block that is being stacked upon
            - ?a - arm: the arm of the robot performing the action
            ```

            ## DISTRACTION TEXT
            """
        )

        types = {
            "arm": "arm for a robot",
            "block": "block that can be stacked and unstacked",
            "table": "table that blocks sits on",
        }

        exp_params = OrderedDict({"?b1": "block", "?b2": "block", "?a": "arm"})

        params, _, _, validation_info = self.domain_builder.formalize_parameters(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            action_name="stack",
            types=types,
            syntax_validator=self.syntax_validator,
        )

        self.assertEqual(exp_params, params)
        self.assertEqual(validation_info[0], True)

    def test_formalize_preconditions(self):
        self.syntax_validator.headers = ["Action Preconditions"]
        self.syntax_validator.error_types = [
            "validate_header",
            "validate_duplicate_headers",
            "validate_unsupported_keywords",
            "validate_params",
        ]

        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT

            ### Action Preconditions
            ```
            (and
                (holding ?a ?b1) ; The arm is holding the top block
                (clear ?b2) ; The bottom block is clear
            )
            ```

            ## DISTRACTION TEXT
            """
        )

        types = {
            "arm": "arm for a robot",
            "block": "block that can be stacked and unstacked",
            "table": "table that blocks sits on",
        }

        params = OrderedDict({"?b1": "block", "?b2": "block", "?a": "arm"})

        exp_preconditions = textwrap.dedent(
            """
            (and
                (holding ?a ?b1) ; The arm is holding the top block
                (clear ?b2) ; The bottom block is clear
            )
            """
        )

        preconditions, _, _, validation_info = (
            self.domain_builder.formalize_preconditions(
                model=self.mock_llm,
                domain_desc="",
                prompt_template="",
                action_name="stack",
                params=params,
                types=types,
                syntax_validator=self.syntax_validator,
                extract_new_preds=False,
            )
        )

        self.assertEqual(exp_preconditions.strip(), preconditions.strip())
        self.assertEqual(validation_info[0], True)

    def test_formalize_effects(self):
        self.syntax_validator.headers = ["Action Effects"]
        self.syntax_validator.error_types = [
            "validate_header",
            "validate_duplicate_headers",
            "validate_unsupported_keywords",
            "validate_params",
        ]

        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT

            ### Action Preconditions
            ```
            (and
                (holding ?a ?b1) ; The arm is holding the top block
                (clear ?b2) ; The bottom block is clear
            )
            ```

            ## DISTRACTION TEXT

            ### Action Effects
            ```
            (and
                (not (holding ?a ?b1)) ; The arm is no longer holding the top block
                (on ?b1 ?b2) ; The top block is now on the bottom block
                (not (clear ?b2)) ; The bottom block is no longer clear
            )
            ```
            """
        )

        types = {
            "arm": "arm for a robot",
            "block": "block that can be stacked and unstacked",
            "table": "table that blocks sits on",
        }

        params = OrderedDict({"?b1": "block", "?b2": "block", "?a": "arm"})

        exp_effects = textwrap.dedent(
            """
            (and
                (not (holding ?a ?b1)) ; The arm is no longer holding the top block
                (on ?b1 ?b2) ; The top block is now on the bottom block
                (not (clear ?b2)) ; The bottom block is no longer clear
            )
            """
        )

        effects, _, _, validation_info = self.domain_builder.formalize_effects(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            action_name="stack",
            params=params,
            types=types,
            syntax_validator=self.syntax_validator,
            extract_new_preds=False,
        )

        self.assertEqual(self.normalize(exp_effects), self.normalize(effects))
        self.assertEqual(validation_info[0], True)

    def test_formalize_predicates(self):
        self.syntax_validator.headers = ["New Predicates"]
        self.syntax_validator.error_types = [
            "validate_types_predicates",
            "validate_format_predicates",
            "validate_duplicate_predicates",
        ]

        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT

            ### New Predicates
            ```
            - (holding ?a - arm ?b - block): true if arm is holding a block
            - (clear ?b - block): true if a block does not have anything on top of it
            ```

            ## DISTRACTION TEXT
            """
        )

        types = {
            "arm": "arm for a robot",
            "block": "block that can be stacked and unstacked",
            "table": "table that blocks sits on",
        }

        predicates = [
            {
                "name": "on",
                "desc": "true if the block ?b1 is on top of the block ?b2",
                "raw": "- (on ?b1 - block ?b2 - block): true if the block ?b1 is on top of the block ?b2",
                "params": OrderedDict({"?b1": "block", "?b2": "block"}),
                "clean": "(on ?b1 - block ?b2 - block)",
            }
        ]

        new_predicates, _, validation_info = self.domain_builder.formalize_predicates(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            types=types,
            predicates=predicates,
            syntax_validator=self.syntax_validator,
        )

        predicates.extend(new_predicates)

        exp_predicates = [
            {
                "name": "holding",
                "desc": "true if arm is holding a block",
                "raw": "- (holding ?a - arm ?b - block): true if arm is holding a block",
                "params": OrderedDict({"?a": "arm", "?b": "block"}),
                "clean": "(holding ?a - arm ?b - block)",
            },
            {
                "name": "clear",
                "desc": "true if a block does not have anything on top of it",
                "raw": "- (clear ?b - block): true if a block does not have anything on top of it",
                "params": OrderedDict({"?b": "block"}),
                "clean": "(clear ?b - block)",
            },
            {
                "name": "on",
                "desc": "true if the block ?b1 is on top of the block ?b2",
                "raw": "- (on ?b1 - block ?b2 - block): true if the block ?b1 is on top of the block ?b2",
                "params": OrderedDict({"?b1": "block", "?b2": "block"}),
                "clean": "(on ?b1 - block ?b2 - block)",
            },
        ]

        self.assertCountEqual(exp_predicates, predicates)
        self.assertEqual(validation_info[0], True)

    def test_formalize_function(self):

        self.syntax_validator.error_types = ["validate_format_functions"]
        self.mock_llm.output = textwrap.dedent(
            """
            ## DISTRACTION TEXT

            ### FUNCTIONS
            ```
            - (battery-level ?r - robot): battery level of robot
            - (humidity ?loc - location): humidity of location
            ```

            ## DISTRACTION TEXT
            """
        )

        types = {"robot": "a robot", "location": "location to be at"}

        functions, _, validation_info = self.domain_builder.formalize_functions(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            types=types,
            syntax_validator=self.syntax_validator,
        )

        exp_functions = [
            {
                "name": "battery-level",
                "desc": "battery level of robot",
                "raw": "- (battery-level ?r - robot): battery level of robot",
                "params": OrderedDict([("?r", "robot")]),
                "clean": "(battery-level ?r - robot)",
            },
            {
                "name": "humidity",
                "desc": "humidity of location",
                "raw": "- (humidity ?loc - location): humidity of location",
                "params": OrderedDict([("?loc", "location")]),
                "clean": "(humidity ?loc - location)",
            },
        ]

        self.assertEqual(validation_info[0], True)
        self.assertEqual(exp_functions, functions)

    def test_formalize_constants(self):
        self.syntax_validator.error_types = ["validate_constant_types"]

        self.mock_llm.output = textwrap.dedent(
            """
            ### DISTRACTION TEXT
            ### CONSTANTS
            ```
            {
                'robot1': 'robot',
                'robot2': 'robot',
                'charging-station': 'location',
                'storage-room': 'location',
                'loading-dock': 'location',
            }
            ```
            ### DISTRACTION TEXT
            """
        )

        types = {"robot": "a robot", "location": "location to be at"}

        exp_constants = {
            "robot1": "robot",
            "robot2": "robot",
            "charging-station": "location",
            "storage-room": "location",
            "loading-dock": "location",
        }

        constants, _, validation_info = self.domain_builder.formalize_constants(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            types=types,
            constants=None,
            syntax_validator=self.syntax_validator,
        )

        self.assertEqual(exp_constants, constants)
        self.assertEqual(validation_info[0], True)

    def test_formalize_domain_level_specs(self):

        self.mock_llm.output = textwrap.dedent(
            """
            ### TYPES
            ```
            [
                {
                    "arm": "'arm for a robot'",
                    "children": []
                },
                {
                    "block": "block that can be stacked and unstacked",
                    "children": [
                        {
                            "heavy_block": "a heavy block that cannot be picked up",
                            "children": []
                        },
                        {
                            "light_block": "a light block that can be picked up",
                            "children": []
                        }
                    ]
                },
                {
                    "table": "table that blocks sits on",
                    "children": []
                }
            ]
            ```
            
            ### CONSTANTS
            ```
            {
                'robot1': 'robot',
                'robot2': 'robot',
                'charging-station': 'location',
                'storage-room': 'location',
                'loading-dock': 'location',
            }
            ```
            
            ### New Predicates
            ```
            - (holding ?a - arm ?b - block): true if arm is holding a block
            - (clear ?b - block): true if a block does not have anything on top of it
            ```
            
            ### FUNCTIONS
            ```
            - (battery-level ?r - robot): battery level of robot
            - (humidity ?loc - location): humidity of location
            ```
            
            """
        )

        results, _, _ = self.domain_builder.formalize_domain_level_specs(
            model=self.mock_llm,
            domain_desc="",
            prompt_template="",
            formalize_types=True,
            formalize_constants=True,
            formalize_predicates=True,
            formalize_functions=True,
        )

        exp_types = [
            {"arm": "'arm for a robot'", "children": []},
            {
                "block": "block that can be stacked and unstacked",
                "children": [
                    {
                        "heavy_block": "a heavy block that cannot be picked up",
                        "children": [],
                    },
                    {
                        "light_block": "a light block that can be picked up",
                        "children": [],
                    },
                ],
            },
            {"table": "table that blocks sits on", "children": []},
        ]

        exp_constants = {
            "robot1": "robot",
            "robot2": "robot",
            "charging-station": "location",
            "storage-room": "location",
            "loading-dock": "location",
        }

        exp_preds = [
            {
                "name": "holding",
                "desc": "true if arm is holding a block",
                "raw": "- (holding ?a - arm ?b - block): true if arm is holding a block",
                "params": OrderedDict([("?a", "arm"), ("?b", "block")]),
                "clean": "(holding ?a - arm ?b - block)",
            },
            {
                "name": "clear",
                "desc": "true if a block does not have anything on top of it",
                "raw": "- (clear ?b - block): true if a block does not have anything on top of it",
                "params": OrderedDict([("?b", "block")]),
                "clean": "(clear ?b - block)",
            },
        ]

        exp_func = [
            {
                "name": "battery-level",
                "desc": "battery level of robot",
                "raw": "- (battery-level ?r - robot): battery level of robot",
                "params": OrderedDict([("?r", "robot")]),
                "clean": "(battery-level ?r - robot)",
            },
            {
                "name": "humidity",
                "desc": "humidity of location",
                "raw": "- (humidity ?loc - location): humidity of location",
                "params": OrderedDict([("?loc", "location")]),
                "clean": "(humidity ?loc - location)",
            },
        ]

        self.assertEqual(results["types"], exp_types)
        self.assertEqual(results["constants"], exp_constants)
        self.assertEqual(results["predicates"], exp_preds)
        self.assertEqual(results["functions"], exp_func)

    def test_generate_domain(self):

        domain = "test_domain"

        types = {"robot": "a robot", "location": "location to be at"}

        constants = {
            "robot1": "robot",
            "robot2": "robot",
            "charging-station": "location",
            "storage-room": "location",
            "loading-dock": "location",
        }

        predicates = [
            {
                "name": "at",
                "desc": "true if robot is at a location",
                "raw": "- (at ?r - robot ?l - location): true if robot is at a location",
                "params": OrderedDict({"?r": "robot", "?l": "location"}),
                "clean": "(at ?r - robot ?l - location)",
            },
            {
                "name": "connected",
                "desc": "true if location ?l1 is connected to location ?l2",
                "raw": "- (connected ?l1 ?l2 - location): true if location ?l1 is connected to location ?l2",
                "params": OrderedDict({"?l1": "location", "?l2": "location"}),
                "clean": "(connected ?l1 ?l2 - location)",
            },
        ]

        functions = [
            {
                "name": "battery-level",
                "desc": "battery level of robot",
                "raw": "(battery-level ?r - robot): battery level of robot",
                "params": OrderedDict([("?r", "robot")]),
                "clean": "(battery-level ?r - robot)",
            },
            {
                "name": "humidity",
                "desc": "humidity of location",
                "raw": "(humidity ?loc - location): humidity of location",
                "params": OrderedDict([("?loc", "location")]),
                "clean": "(humidity ?loc - location)",
            },
        ]

        actions = [
            {
                "name": "move",
                "params": {"?r": "robot", "?l1": "location", "?l2": "location"},
                "preconditions": "(and (at ?r ?l1) (connected ?l1 ?l2))",
                "effects": "(and (not (at ?r ?l1)) (at ?r ?l2))",
            },
            {
                "name": "pick",
                "params": {"?r": "robot", "?l": "location"},
                "preconditions": "(and (at ?r ?l) (not (holding ?r)))",
                "effects": "(holding ?r)",
            },
        ]

        requirements = [":strips", ":typing", ":numeric-fluents"]

        expected_output = textwrap.dedent(
            """
            (define (domain test_domain)
                (:requirements
                    :negative-preconditions :numeric-fluents :strips :typing)

                (:types 
                    location 
                    robot
                )
                
                (:constants 
                    robot1 - robot
                    robot2 - robot
                    charging-station - location
                    storage-room - location
                    loading-dock - location
                )

                (:predicates 
                    (at ?r - robot ?l - location)
                    (connected ?l1 ?l2 - location)
                )
                
                (:functions 
                    (battery-level ?r - robot)
                    (humidity ?loc - location)
                )

                (:action move
                    :parameters (
                        ?r - robot ?l1 ?l2 - location
                    )
                    :precondition
                        (and (at ?r ?l1) (connected ?l1 ?l2))
                    :effect
                        (and (not (at ?r ?l1)) (at ?r ?l2))
                )

                (:action pick
                    :parameters (
                        ?r - robot ?l - location
                    )
                    :precondition
                        (and (at ?r ?l) (not (holding ?r)))
                    :effect
                        (holding ?r)
                )
            )
            """
        )

        result = self.domain_builder.generate_domain(
            domain_name=domain,
            types=types,
            constants=constants,
            predicates=predicates,
            functions=functions,
            actions=actions,
            # requirements=requirements,
        )

        self.assertEqual(self.normalize(result), self.normalize(expected_output))


if __name__ == "__main__":
    unittest.main()



"""
L2P is compatible with FastDownward: https://www.fast-downward.org

For usage, users must clone or download the submodule /downward separately and direct the
`planner_path` to the folder. This module is not necessary to use L2P, but for ease of use
to produce plans from generated domain and problem PDDL specifications via LLMs.
"""

import subprocess, re

# Define the exit codes
SUCCESS = 0
SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY = 1
SEARCH_PLAN_FOUND_AND_OUT_OF_TIME = 2
SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY_AND_TIME = 3

TRANSLATE_UNSOLVABLE = 10
SEARCH_UNSOLVABLE = 11
SEARCH_UNSOLVED_INCOMPLETE = 12

TRANSLATE_OUT_OF_MEMORY = 20
TRANSLATE_OUT_OF_TIME = 21
SEARCH_OUT_OF_MEMORY = 22
SEARCH_OUT_OF_TIME = 23
SEARCH_OUT_OF_MEMORY_AND_TIME = 24

TRANSLATE_CRITICAL_ERROR = 30
TRANSLATE_INPUT_ERROR = 31
SEARCH_CRITICAL_ERROR = 32
SEARCH_INPUT_ERROR = 33
SEARCH_UNSUPPORTED = 34
DRIVER_CRITICAL_ERROR = 35
DRIVER_INPUT_ERROR = 36
DRIVER_UNSUPPORTED = 37


class FastDownward:

    def __init__(self, planner_path: str):
        self.planner_path = planner_path  # directory of FastDownward planner

    def run_fast_downward(
        self, domain_file: str, problem_file: str, search_alg: str = "lama-first"
    ):
        """
        Main function to run planner.

        Args:
            domain_file (str): PDDL domain file path
            problem_file (str): PDDL problem file path
            search_alg (str): search algorithm/heuristic to use
                + refer to: https://www.fast-downward.org/PlannerUsage

        Returns:
            success (bool): if a plan was found, otherwise False for incomplete.
            plan_output (str): plan output information.

        """
        try:
            result = subprocess.run(
                [self.planner_path, "--alias", search_alg, domain_file, problem_file],
                capture_output=True,
                text=True,
            )

            exitcodes = [result.returncode]

            if result.returncode == SUCCESS:
                # Planning succeeded
                print("Planning succeeded!")
                print(
                    "All run components successfully terminated (translator: completed, search: found a plan, validate: validated a plan)"
                )

                # Extract the plan steps from the output
                plan_output = self.extract_plan_steps(result.stdout)
                if plan_output:
                    return True, plan_output
                else:
                    return False, "No plan found in the output."
            else:
                # Planning failed
                exitcode, plan_found = self.generate_portfolio_exitcode(exitcodes)
                return False, self.handle_error(exitcode, plan_found)
        except Exception as e:
            print("An error occurred while running the planner.")
            return False, str(e)

    def extract_plan_steps(self, output):
        plan_steps = re.findall(r"^\w+.*\(.*\)", output, re.MULTILINE)
        return "\n".join(plan_steps)

    def handle_error(self, exitcode, plan_found):
        if plan_found:
            if exitcode == SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY:
                return "Plan found but the search ran out of memory."
            elif exitcode == SEARCH_PLAN_FOUND_AND_OUT_OF_TIME:
                return "Plan found but the search ran out of time."
            elif exitcode == SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY_AND_TIME:
                return "Plan found but the search ran out of memory and time."
            else:
                return f"Unknown plan occurred with exit code: {exitcode}"
        else:
            if exitcode == TRANSLATE_UNSOLVABLE:
                return "Translate phase determined the problem is unsolvable."
            elif exitcode == SEARCH_UNSOLVABLE:
                return "Search phase determined the problem is unsolvable."
            elif exitcode == SEARCH_UNSOLVED_INCOMPLETE:
                return "Search phase was incomplete and did not solve the problem."
            elif exitcode == TRANSLATE_OUT_OF_MEMORY:
                return "Translate phase ran out of memory."
            elif exitcode == TRANSLATE_OUT_OF_TIME:
                return "Translate phase ran out of time."
            elif exitcode == SEARCH_OUT_OF_MEMORY:
                return "Search phase ran out of memory."
            elif exitcode == SEARCH_OUT_OF_TIME:
                return "Search phase ran out of time."
            elif exitcode == SEARCH_OUT_OF_MEMORY_AND_TIME:
                return "Search phase ran out of memory and time."
            elif exitcode == TRANSLATE_CRITICAL_ERROR:
                return "Critical error in translate phase."
            elif exitcode == TRANSLATE_INPUT_ERROR:
                return "Input error in translate phase."
            elif exitcode == SEARCH_CRITICAL_ERROR:
                return "Critical error in search phase."
            elif exitcode == SEARCH_INPUT_ERROR:
                return "Input error in search phase."
            elif exitcode == SEARCH_UNSUPPORTED:
                return "Search phase encountered an unsupported feature."
            elif exitcode == DRIVER_CRITICAL_ERROR:
                return "Critical error in the driver."
            elif exitcode == DRIVER_INPUT_ERROR:
                return "Input error in the driver."
            elif exitcode == DRIVER_UNSUPPORTED:
                return "Driver encountered an unsupported feature."
            else:
                return f"Unknown error occurred with exit code: {exitcode}"

    def is_unrecoverable(self, exitcode):
        # Exit codes in the range from 30 to 39 represent unrecoverable failures.
        return 30 <= exitcode < 40

    def generate_portfolio_exitcode(self, exitcodes):

        print("Exit codes: {}".format(exitcodes))
        exitcodes = set(exitcodes)
        unrecoverable_codes = [
            code for code in exitcodes if self.is_unrecoverable(code)
        ]

        # There are unrecoverable exit codes.
        if unrecoverable_codes:
            print("Error: Unexpected exit codes: {}".format(unrecoverable_codes))
            if len(unrecoverable_codes) == 1:
                return (unrecoverable_codes[0], False)
            else:
                return (SEARCH_CRITICAL_ERROR, False)

        # At least one plan was found.
        if SUCCESS in exitcodes:
            if SEARCH_OUT_OF_MEMORY in exitcodes and SEARCH_OUT_OF_TIME in exitcodes:
                return (SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY_AND_TIME, True)
            elif SEARCH_OUT_OF_MEMORY in exitcodes:
                return (SEARCH_PLAN_FOUND_AND_OUT_OF_MEMORY, True)
            elif SEARCH_OUT_OF_TIME in exitcodes:
                return (SEARCH_PLAN_FOUND_AND_OUT_OF_TIME, True)
            else:
                return (SUCCESS, True)

        # A config proved unsolvability or did not find a plan.
        for code in [SEARCH_UNSOLVABLE, SEARCH_UNSOLVED_INCOMPLETE]:
            if code in exitcodes:
                return (code, False)

        # No plan was found due to hitting resource limits.
        if SEARCH_OUT_OF_MEMORY in exitcodes and SEARCH_OUT_OF_TIME in exitcodes:
            return (SEARCH_OUT_OF_MEMORY_AND_TIME, False)
        elif SEARCH_OUT_OF_MEMORY in exitcodes:
            return (SEARCH_OUT_OF_MEMORY, False)
        elif SEARCH_OUT_OF_TIME in exitcodes:
            return (SEARCH_OUT_OF_TIME, False)

        assert False, "Error: Unhandled exit codes: {}".format(exitcodes)



"""
This module contains collection of functions for formatting Python PDDL components into PDDL format strings.
"""

import json
import re
from collections import defaultdict, OrderedDict
from .pddl_types import Action, Function, Predicate


# ---- PDDL DOMAIN FORMATTER ----


def format_types(
    types: dict[str, str] | list[dict[str, str]] | None = None,
) -> dict[str, str]:
    """
    Formats nested Python type hierarchies into flat dictionaries. Flat type dictionaries
    (no hierarchies) are returned as default.
    """

    if not types:
        return None

    result = {}

    def is_nested_format(typ) -> bool:
        if isinstance(typ, list):
            return True
        return any(isinstance(v, list) and k == "children" for k, v in typ.items())

    def process_node(node, parent_type=None):
        if not isinstance(node, dict):
            return

        type_name = next((k for k in node if k != "children"), None)
        if type_name is None:
            return

        description = node[type_name]
        parent = parent_type if parent_type else type_name

        name = f"{type_name} - {parent}" if type_name != parent else type_name
        result[name] = f"; {description}" if description else ""

        for child in node.get("children", []):
            process_node(child, type_name)

    if is_nested_format(types):
        if isinstance(types, list):
            for node in types:
                process_node(node)
        else:
            process_node(types)
    else:
        for type_name, description in types.items():
            result[type_name] = f"; {description}" if description else ""

    return result


def format_types_to_string(types: dict[str, str] | list[dict[str, str]]) -> str:
    """Formats a type hierarchy (flat or nested) into a PDDL-style string."""
    formatted = format_types(types)

    # appends `object` to parent type (required for some PDDL parsers)
    type_groups = {}
    for type_name, desc in formatted.items():

        # if type has a parent
        if " - " in type_name:
            type_part, parent_part = type_name.split(" - ")
            parent = parent_part.strip()
            type_part = type_part.strip()
        # if type does not have a parent
        else:
            # requests if object is appended
            parent = type_name.strip()
            type_part = None

        # if there is a parent (could be super type or object)
        if parent:
            # if unique parent, add it to type groups
            if parent not in type_groups:
                type_groups[parent] = []
            if type_part:
                type_groups[parent].append(type_part)

    # build the output lines
    lines = []

    # handle top-level objects first
    if "object" in type_groups:
        top_level_types = sorted(t for t in type_groups["object"] if t != "object")
        if top_level_types:
            lines.append(f"{' '.join(top_level_types)} - object")
        del type_groups["object"]

    # handle child groups
    for parent, child_types in sorted(type_groups.items()):

        if child_types:
            child_line = " ".join(sorted(child_types))
            lines.append(f"{child_line} - {parent}")
        else:
            lines.append(f"{parent}")

    return "\n".join(lines)


def format_constants(constants: dict[str, str]) -> str:
    """Formats constant dictionary into PDDL-style string"""
    lines = []
    for const_name, const_type in constants.items():
        lines.append(f"{const_name} - {const_type}")

    return "\n".join(lines)


def format_expression(expressions: list[Predicate] | list[Function]) -> str:
    """Formats predicate/function list into a PDDL-style string, removing exact duplicates."""
    unique = dict()  # key = (name.lower(), tuple(params)), value = clean string
    for exp in expressions:
        key = (exp["name"], tuple(exp["params"]))
        if key not in unique:
            unique[key] = exp["clean"].replace(":", " ; ")

    return "\n".join(unique.values())


def format_params(parameters: OrderedDict) -> str:
    """Helper function to format parameters (as its type) into a PDDL-style string."""
    grouped_params = defaultdict(list)
    for name, type_ in parameters.items():
        # ensure name starts with '?'
        name = name if name.startswith("?") else f"?{name}"
        clean_type = type_.strip() if type_ else None
        grouped_params[clean_type].append(name)

    # separate out typed and untyped/object parameters
    typed_parts = []
    untyped_parts = []
    for type_, names in grouped_params.items():
        if type_ is None or type_ == "object":
            untyped_parts.append(" ".join(names))  # untyped parameters
        else:
            typed_parts.append(f"{' '.join(names)} - {type_}")

    return " ".join(typed_parts + untyped_parts)


def format_actions(actions: list[Action]) -> str:
    """Formats a list of Actions into a PDDL-style string."""
    desc = ""
    for action in actions:
        desc += "\n\n" + indent(format_action_desc(action), level=1)
    return desc


def format_action_desc(action: Action) -> str:
    """Helper function to format individual action descriptions"""
    param_str = format_params(action["params"])

    preconditions = "\n".join(
        line for line in action["preconditions"].splitlines() if line.strip()
    )

    effects = "\n".join(line for line in action["effects"].splitlines() if line.strip())

    desc = f"(:action {action['name']}\n"
    desc += f"   :parameters (\n{indent(string=param_str, level=2)}\n   )\n"
    desc += f"   :precondition\n{indent(string=preconditions, level=2)}\n"
    desc += f"   :effect\n{indent(string=effects, level=2)}\n"
    desc += ")"
    return desc


# ---- PDDL PROBLEM FORMATTER ----


def format_objects(objects: dict[str, str]) -> str:
    """Formats task objects into a PDDL-style string."""
    objects = "\n".join(
        [f"{obj} - {type}" if type else f"{obj}" for obj, type in objects.items()]
    )
    return objects


def format_initial(initial_states: list[dict[str, str]]) -> str:
    """Formats task initial states into a PDDL-style string."""

    full_str = []

    for state in initial_states:
        # if function statement
        if state.get("func_name"):
            state_op, state_func, state_params, state_value = (
                state["op"],
                state["func_name"],
                " ".join(state["params"]),
                state["value"],
            )
            full_str.append(f"({state_op} ({state_func} {state_params}) {state_value})")
        # if predicate statement
        elif state.get("pred_name"):

            inner_str = f"({state['pred_name']} {' '.join(state['params'])})"

            full_str.append(f"(not {inner_str})" if state["neg"] else inner_str)

    initial_states_str = "\n".join(full_str)  # combine the states into a single string

    return initial_states_str


def format_goal(goal_states: list[dict[str, str]]) -> str:
    """Formats task goal states into a PDDL-style string."""
    full_str = []

    for state in goal_states:
        # if function statement
        if state.get("func_name"):
            state_op, state_func, state_params, state_value = (
                state["op"],
                state["func_name"],
                " ".join(state["params"]),
                state["value"],
            )
            full_str.append(f"({state_op} ({state_func} {state_params}) {state_value})")
        # if predicate statement
        elif state.get("pred_name"):
            inner_str = f"({state['pred_name']} {' '.join(state['params'])})"
            full_str.append(f"(not {inner_str})" if state["neg"] else inner_str)

        goal_states_str = "\n".join(full_str)  # combine the states into a single string

    goal_states_str = f"(and \n{indent(goal_states_str, 1)}\n)"

    return goal_states_str


# ---- HELPER FUNCTIONS ----


def indent(string: str, level: int = 2):
    """Indent string helper function to format PDDL domain/task"""
    return "   " * level + string.replace("\n", f"\n{'   ' * level}")


def remove_comments(text: str, comment_prefixes=[";", "#", "//"]) -> str:
    """Remove comments from text using multiple prefix styles."""

    lines = text.splitlines()
    cleaned_lines = []

    for line in lines:
        stripped_line = line
        for prefix in comment_prefixes:
            if prefix in stripped_line:
                # only remove comment if prefix is not inside quotes or code
                stripped_line = stripped_line.split(prefix, 1)[0]
        cleaned_lines.append(stripped_line.rstrip())

    # remove blank lines and normalize whitespace
    cleaned = "\n".join(line for line in cleaned_lines if line.strip())
    cleaned = re.sub(r"\n{2,}", "\n\n", cleaned)  # collapse multiple newlines

    return cleaned


def format_pddl_expr(expr):
    """Parses nested list format into bracket."""
    if isinstance(expr, list):
        return "(" + " ".join(format_pddl_expr(e) for e in expr) + ")"
    else:
        return str(expr)


def pretty_print_dict(data):
    """Formats dictionary or list of dictionaries in JSON format for readability."""
    if isinstance(data, (dict, list)):
        return json.dumps(data, indent=4)
    else:
        raise TypeError("Input must be a dictionary or a list of dictionaries")


def pretty_print_expression(expressions: list[Predicate] | list[Function]) -> str:
    """Formats list of predicates easier for readability"""
    if not expressions:
        return ""
    return "\n".join(
        f"{i + 1}. {exp['name']}: {exp.get('desc', 'No description provided') or 'No description provided'}"
        for i, exp in enumerate(expressions)
    )



"""
PDDL Problem Formalization/Generation Functions

This module defines the `TaskBuilder` class and related utilities for constructing
PDDL problem specifications programatically.

Refer to: https://marcustantakoun.github.io/l2p.github.io/l2p.html for more information
how to use class functions. Refer to /templates in: https://github.com/AI-Planning/l2p
for how to structurally prompt LLMs so they are compatible with class function parsing.
"""

import time
from .llm import BaseLLM, require_llm
from .utils import *


class TaskBuilder:
    def __init__(
        self,
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
    ) -> None:
        """
        Initializes an L2P task builder object.

        Args:
            objects (dict[str,str]): current dictionary of task objects in specification
            initial (list[dict[str,str]]): current initial states in specification
            goal (list[dict[str,str]]): current goal states in specification
        """

        self.objects = objects or {}
        self.initial = initial or []
        self.goal = goal or []

    """Formalize/generate functions"""

    @require_llm
    def formalize_objects(
        self,
        model: BaseLLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] | list[dict[str, str]] | None = None,
        constants: dict[str, str] | None = None,
        syntax_validator: SyntaxValidator = None,
        max_retries: int = 3,
    ) -> tuple[dict[str, str], str, tuple[bool, str]]:
        """
        Formalizes PDDL :objects via LLM.

        Args:
            model (BaseLLM): LLM to query
            problem_desc (str): general problem description
            prompt_template (str): structured prompt template for :objects extraction
            types (dict[str,str] | list[dict[str,str]]): current types in specification, defaults to None
            constants (dict[str,str]): current constants in specification, defaults to None
            syntax_validator (SyntaxValidator): syntax checker for generated objects, defaults to None
            max_retries (int): max # of retries if failure occurs

        Returns:
            objects (dict[str,str]): dictionary of object types {<name>: <type>}
            llm_output (str): the raw string BaseLLM response
            validation_info (tuple[bool,str]): validation info containing pass flag and error message
        """

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )

        prompt = (
            prompt_template.replace("{problem_desc}", problem_desc)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
        )

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()
                llm_output = model.query(prompt=prompt)  # get BaseLLM response

                # extract respective types from response
                objects = parse_objects(llm_output=llm_output)

                # run syntax validation if applicable
                validation_info = (True, "All validations passed.")
                if syntax_validator:
                    for error_type in syntax_validator.error_types:
                        validator = getattr(syntax_validator, f"{error_type}", None)
                        if not callable(validator):
                            continue

                        # dispatch based on expected arguments
                        if error_type == "validate_header":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_duplicate_headers":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_unsupported_keywords":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_task_objects":
                            validation_info = validator(objects, types)

                        if not validation_info[0]:
                            return objects, llm_output, validation_info

                return objects, llm_output, validation_info

            except Exception as e:
                print(
                    f"Error on attempt {attempt + 1}/{max_retries}: {e}\n"
                    f"LLM Output:\n{llm_output if 'llm_output' in locals() else 'None'}\nRetrying...\n"
                )
                time.sleep(2)

        raise RuntimeError("Max retries exceeded. Failed to extract objects.")

    @require_llm
    def formalize_initial_state(
        self,
        model: BaseLLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] | list[dict[str, str]] | None = None,
        constants: dict[str, str] | None = None,
        predicates: list[Predicate] | None = None,
        functions: list[Function] | None = None,
        objects: dict[str, str] | None = None,
        initial: list[dict[str, str]] | None = None,
        goal: list[dict[str, str]] | None = None,
        syntax_validator: SyntaxValidator = None,
        max_retries: int = 3,
    ) -> tuple[list[dict[str, str]], str, tuple[bool, str]]:
        """
        Formalizes PDDL :init states via LLM.

        Args:
            model (BaseLLM): LLM to query
            problem_desc (str): general problem description
            prompt_template (str): structured prompt template for :init extraction
            types (dict[str,str] | list[dict[str,str]]): current types in domain, defaults to None
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of current predicates in domain, defaults to None
            functions (list[Function]): list of current functions in specification, defaults to None
            objects (dict[str,str]): current dictionary of task :objects in specification, defaults to None
            initial (list[dict[str,str]]): current :init states in specification, defaults to None
            goal (list[dict[str,str]]): current :goal states in specification, defaults to None
            syntax_validator (SyntaxValidator): syntax checker for generated initial states, defaults to None
            max_retries (int): max # of retries if failure occurs

        Returns:
            initial (list[dict[str,str]]): list of dictionaries containing initial states consisting of:
                {<predicate_name>:<str>, <params>:<list[str]>, <neg>:<bool>} OR
                {<function_name>:<str>, <params>:<list[str]>, <value>:<int>, <op>:<str>}
            llm_output (str): the raw string BaseLLM response
            validation_info (tuple[bool,str]): validation info containing pass flag and error message
        """

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )
        obj_str = format_objects(objects) if objects else "No objects provided."
        init_str = format_initial(initial) if initial else "No initial state provided."
        goal_str = format_goal(goal) if goal else "No goal state provided."

        prompt = (
            prompt_template.replace("{problem_desc}", problem_desc)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{objects}", obj_str)
            .replace("{initial_state}", init_str)
            .replace("{goal_state}", goal_str)
        )

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()
                llm_output = model.query(prompt=prompt)

                # extract respective types from response
                initial = parse_initial(llm_output=llm_output)

                # run syntax validation if applicable
                validation_info = (True, "All validations passed.")
                if syntax_validator:
                    for error_type in syntax_validator.error_types:
                        validator = getattr(syntax_validator, f"{error_type}", None)
                        if not callable(validator):
                            continue

                        # dispatch based on expected arguments
                        if error_type == "validate_header":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_duplicate_headers":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_unsupported_keywords":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_task_states":
                            validation_info = validator(
                                initial, objects, predicates, "initial"
                            )

                        if not validation_info[0]:
                            return initial, llm_output, validation_info

                return initial, llm_output, validation_info

            except Exception as e:
                print(
                    f"Error on attempt {attempt + 1}/{max_retries}: {e}\n"
                    f"LLM Output:\n{llm_output if 'llm_output' in locals() else 'None'}\nRetrying...\n"
                )
                time.sleep(2)

        raise RuntimeError("Max retries exceeded. Failed to extract initial states.")

    @require_llm
    def formalize_goal_state(
        self,
        model: BaseLLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] | list[dict[str, str]] | None = None,
        constants: dict[str, str] | None = None,
        predicates: list[Predicate] | None = None,
        functions: list[Function] | None = None,
        objects: dict[str, str] | None = None,
        initial: list[dict[str, str]] | None = None,
        goal: list[dict[str, str]] | None = None,
        syntax_validator: SyntaxValidator = None,
        max_retries: int = 3,
    ) -> tuple[list[dict[str, str]], str, tuple[bool, str]]:
        """
        Formalizes PDDL :goal states via LLM.

        Args:
            model (BaseLLM): LLM to query
            problem_desc (str): general problem description
            prompt_template (str): structured prompt template for :goal extraction
            types (dict[str,str] | list[dict[str,str]]): current :types in domain, defaults to None
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of current predicates in domain, defaults to None
            functions (list[Function]): list of current functions in specification, defaults to None
            objects (dict[str,str]): current dictionary of task :objects in specification, defaults to None
            initial (list[dict[str,str]]): current :init states in specification, defaults to None
            goal (list[dict[str,str]]): current :goal states in specification, defaults to None
            syntax_validator (SyntaxValidator): syntax checker for generated goal states, defaults to None
            max_retries (int): max # of retries if failure occurs

        Returns:
            goal (list[dict[str,str]]): list of dictionaries containing goal states consisting of:
                {<predicate_name>:<str>, <params>:<list[str]>, <neg>:<bool>} OR
                {<function_name>:<str>, <params>:<list[str]>, <value>:<int>, <op>:<str>}
            llm_output (str): the raw string BaseLLM response
            validation_info (tuple[bool,str]): validation info containing pass flag and error message
        """

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )
        obj_str = format_objects(objects) if objects else "No objects provided."
        init_str = format_initial(initial) if initial else "No initial state provided."
        goal_str = format_goal(goal) if goal else "No goal state provided."

        prompt = (
            prompt_template.replace("{problem_desc}", problem_desc)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{objects}", obj_str)
            .replace("{initial_state}", init_str)
            .replace("{goal_state}", goal_str)
        )

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()
                llm_output = model.query(prompt=prompt)

                # extract respective types from response
                goal = parse_goal(llm_output=llm_output)

                # run syntax validation if applicable
                validation_info = (True, "All validations passed.")
                if syntax_validator:
                    for error_type in syntax_validator.error_types:
                        validator = getattr(syntax_validator, f"{error_type}", None)
                        if not callable(validator):
                            continue

                        # dispatch based on expected arguments
                        if error_type == "validate_header":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_duplicate_headers":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_unsupported_keywords":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_task_states":
                            validation_info = validator(
                                goal, objects, predicates, "goal"
                            )

                        if not validation_info[0]:
                            return goal, llm_output, validation_info

                return goal, llm_output, validation_info

            except Exception as e:
                print(
                    f"Error on attempt {attempt + 1}/{max_retries}: {e}\n"
                    f"LLM Output:\n{llm_output if 'llm_output' in locals() else 'None'}\nRetrying...\n"
                )
                time.sleep(2)

        raise RuntimeError("Max retries exceeded. Failed to extract goal states.")

    @require_llm
    def formalize_task(
        self,
        model: BaseLLM,
        problem_desc: str,
        prompt_template: str,
        types: dict[str, str] | list[dict[str, str]] | None = None,
        constants: dict[str, str] | None = None,
        predicates: list[Predicate] | None = None,
        functions: list[Function] | None = None,
        syntax_validator: SyntaxValidator = None,
        max_retries: int = 3,
    ) -> tuple[
        dict[str, str],
        list[dict[str, str]],
        list[dict[str, str]],
        str,
        tuple[bool, str],
    ]:
        """
        Formalizes whole task specification via LLM.

        Args:
            model (BaseLLM): LLM to query
            problem_desc (str): general problem description
            prompt_template (str): structured prompt template for :problem extraction
            types (dict[str,str] | list[dict[str,str]]): current :types in domain, defaults to None
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of current predicates in domain, defaults to None
            functions (list[Function]): list of current functions in specification, defaults to None
            syntax_validator (SyntaxValidator): syntax checker for generated :problem, defaults to None
            max_retries (int): max # of retries if failure occurs

        Returns:
            objects (dict[str,str]): dictionary of object names and assigned types
            initial (list[dict[str,str]]): list of dictionary of initial states
            goal (list[dict[str,str]]): list of dictionary of goal states
            llm_output (str): the raw string BaseLLM response
            validation_info (tuple[bool,str]): validation info containing pass flag and error message
        """

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            prompt_template.replace("{problem_desc}", problem_desc)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
        )

        # iterate through attempts in case of extraction failure
        for attempt in range(max_retries):
            try:
                model.reset_tokens()
                llm_output = model.query(prompt=prompt)

                # extract respective types from response
                objects = parse_objects(llm_output=llm_output)
                initial = parse_initial(llm_output=llm_output)
                goal = parse_goal(llm_output=llm_output)

                # run syntax validation if applicable
                validation_info = (True, "All validations passed.")
                if syntax_validator:
                    for error_type in syntax_validator.error_types:
                        validator = getattr(syntax_validator, f"{error_type}", None)
                        if not callable(validator):
                            continue

                        # dispatch based on expected arguments
                        if error_type == "validate_header":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_duplicate_headers":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_unsupported_keywords":
                            validation_info = validator(llm_output)
                        elif error_type == "validate_task_objects":
                            validation_info = validator(objects, types)
                        elif error_type == "validate_task_states":
                            validation_info = validator(
                                states=initial,
                                objects=objects,
                                predicates=predicates,
                                functions=functions,
                                state_type="initial",
                            )
                            if validation_info[0]:
                                validation_info = validator(
                                    states=goal,
                                    objects=objects,
                                    predicates=predicates,
                                    functions=functions,
                                    state_type="goal",
                                )

                        if not validation_info[0]:
                            return objects, initial, goal, llm_output, validation_info

                return objects, initial, goal, llm_output, validation_info

            except Exception as e:
                print(
                    f"Error on attempt {attempt + 1}/{max_retries}: {e}\n"
                    f"LLM Output:\n{llm_output if 'llm_output' in locals() else 'None'}\nRetrying...\n"
                )
                time.sleep(2)

        raise RuntimeError("Max retries exceeded. Failed to extract task.")

    """Delete functions"""

    def delete_objects(self, object: dict[str, str]):
        """Deletes specific item in :objects from current specification"""
        if self.objects is not None:
            self.objects = {
                var: type_ for var, type_ in self.objects.items() if var != object
            }

    def delete_initial_state(self, state: dict[str, str]):
        """Deletes specific :init state from current specification"""
        if self.initial is not None:
            self.initial = [s for s in self.initial if s != state]

    def delete_goal_state(self, state: dict[str, str]):
        """Deletes specific PDDL :goal state from current specification"""
        if self.goal is not None:
            self.goal = [s for s in self.goal if s != state]

    """Set functions"""

    def set_objects(self, objects: dict[str, str]):
        """Sets PDDL :objects for current specification"""
        self.objects = objects

    def set_initial(self, initial: list[dict[str, str]]):
        """Sets PDDL :init states for current specification"""
        self.initial = initial

    def set_goal(self, goal: list[dict[str, str]]):
        """Sets PDDL :goal states for current specification"""
        self.goal = goal

    """Get functions"""

    def get_objects(self) -> dict[str, str]:
        """Returns PDDL :objects from current specification"""
        return self.objects

    def get_initial(self) -> list[dict[str, str]]:
        """Returns PDDL :init states from current specification"""
        return self.initial

    def get_goal(self) -> list[dict[str, str]]:
        """Returns PDDL :goal states from current specification"""
        return self.goal

    def generate_task(
        self,
        domain_name: str,
        problem_name: str,
        objects: dict[str, str],
        initial: list[dict[str, str]],
        goal: list[dict[str, str]],
    ) -> str:
        """
        Generates PDDL problem from given information.

        Args:
            domain_name (str): domain name
            problem_name (str): specific task instance name
            objects (dict[str,str]): PDDL :objects
            initial (list[dict[str,str]]): PDDL :init states
            goal (list[dict[str,str]]): PDDL :goal states

        Returns:
            desc (str): PDDL problem in string format
        """

        desc = "(define\n"
        desc += f"   (problem {problem_name})\n"
        desc += f"   (:domain {domain_name})\n\n"
        desc += f"   (:objects \n{indent(format_objects(objects))}\n   )\n\n"
        desc += f"   (:init\n{indent(format_initial(initial))}\n   )\n\n"
        desc += f"   (:goal\n{indent(format_goal(goal))}\n   )\n"
        desc += ")"
        desc = desc.replace("AND", "and").replace("OR", "or")
        return desc



"""
PDDL Feedback Generation Utilities

This module provides tools for generating feedback on LLM-generated PDDL domains and tasks.
It supports multiple feedback strategies including human-written or LLM-generated feedback.

NOTE: is worth noting that the usefulness of LLM feedback is experimental. It is inspired by the NL2PLAN framework
    (Gestrin et al., 2024) and is designed to provide feedback to LLM output w/o human intervention.
"""

from collections import OrderedDict
from .utils import *
from .llm import BaseLLM, require_llm


class FeedbackBuilder:
    """
    NOTE: this class only returns feedback flag/messages from original LLM output. Users
        must provide their own implementation of using this feedback to revise outputs.
    """

    @require_llm
    def get_feedback(
        self,
        model: BaseLLM,
        feedback_template: str,
        feedback_type: str,
        llm_output: str,
    ) -> tuple[bool, str]:
        """
        This retrieves the type of feedback user requests and returns feedack message.
        feedback_type takes in either "human" or "llm"
        """

        if feedback_type.lower() == "human":
            feedback_msg = self.human_feedback(llm_output)
        elif feedback_type.lower() == "llm":
            model.reset_tokens()
            feedback_msg = model.query(prompt=feedback_template)
        else:
            raise ValueError("Invalid feedback_type. Expected 'human' or 'llm'")

        # retrieve feedback message
        no_feedback = self.feedback_state(info=feedback_msg)

        if no_feedback:
            return True, feedback_msg

        return False, feedback_msg

    def feedback_state(self, info: str):
        """
        Confirms if feedback is needed from output.

        It expects a `### JUDGMENT` header with enclosed (```) boxes with an answer
        of either 'no feedback' or content of suggestion.

        Refer to l2p/templates/feedback_templates for proper LLM output structuring.
        """
        try:
            judgement_head = parse_heading(info, "JUDGMENT")
            judgement_raw = combine_blocks(judgement_head)
            if "no feedback" in judgement_raw.lower():
                return True
            else:
                return False
        except ValueError as e:
            # if heading is not found no feedback is needed
            return True

    def human_feedback(self, info: str):
        """This enables human-in-the-loop feedback mechanism."""

        print("START OF INFO\n", info)
        print("\nEND OF INFO\n\n")
        contents = []
        print("Provide feedback (or type 'done' to finish):\n")
        while True:
            line = input()
            if line.strip().lower() == "done":
                break
            contents.append(line)
        response = "\n".join(contents)

        if response.lower() == "no feedback" or response == "":
            feedback = "### JUDGMENT\n```\nno feedback\n```"
        else:
            feedback = f"### JUDGMENT\n```\n{response}\n```"

        return feedback

    @require_llm
    def type_feedback(
        self,
        model: BaseLLM,
        domain_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        types: dict[str, str] | list[dict[str, str]] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output for :types.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            types (dict[str,str] | list[dict[str,str]]): PDDL types of current specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        types_str = pretty_print_dict(types) if types else "No types provided."

        prompt = (
            feedback_template.replace("{types}", types_str)
            .replace("{domain_desc}", domain_desc)
            .replace("{llm_output}", llm_output)
        )

        # retrieve feedback for types
        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def nl_action_feedback(
        self,
        model: BaseLLM,
        domain_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        types: dict[str, str] | list[dict[str, str]] = None,
        nl_actions: dict[str, str] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output for list of natural language actions for the domain

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            types (dict[str,str] | list[dict[str,str]]): PDDL types of current specification
            nl_actions (dict[str,str]): optional to supplement feedback prompt

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        types_str = pretty_print_dict(types) if types else "No types provided."
        nl_act_str = (
            pretty_print_dict(nl_actions) if nl_actions else "No actions provided."
        )

        prompt = (
            feedback_template.replace("{domain_desc}", domain_desc)
            .replace("{types}", types_str)
            .replace("{nl_actions}", nl_act_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def pddl_action_feedback(
        self,
        model: BaseLLM,
        domain_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        action: Action = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
        functions: list[Function] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of a PDDL action.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            action (Action): current action specifications
            types (dict[str,str] | list[dict[str,str]]): PDDL types of current specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): PDDL predicates of current specification
            functions (list[Function]): PDDL functions of current specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        act_name_str = action["name"] if action else "No action name provided."
        params_str = (
            "\n".join([f"{name} - {type}" if type else f"{name}" for name, type in action["params"].items()])
            if action
            else "No parameters provided"
        )
        prec_str = action["preconditions"] if action else "No preconditions provided."
        eff_str = action["effects"] if action else "No effects provided."

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            feedback_template.replace("{domain_desc}", domain_desc)
            .replace("{action_name}", act_name_str)
            .replace("{action_params}", params_str)
            .replace("{action_preconditions}", prec_str)
            .replace("{action_effects}", eff_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def parameter_feedback(
        self,
        model: BaseLLM,
        domain_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        parameter: OrderedDict = None,
        action_name: str = None,
        action_desc: str = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of a PDDL action parameter.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            parameter (OrderedDict): PDDL params of current action
            action_name (str): name of action
            action_desc (str): description of action
            types (dict[str,str] | list[dict[str,str]]): PDDL types of current specification
            constants (dict[str,str]): current constants in specification, defaults to None

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        act_name_str = action_name if action_name else "No action name provided"
        act_desc_str = action_desc if action_desc else "No action description provided"
        params_str = (
            "\n".join([f"{name} - {type}" for name, type in parameter.items()])
            if parameter
            else "No parameters provided"
        )

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )

        prompt = (
            feedback_template.replace("{domain_desc}", domain_desc)
            .replace("{action_name}", act_name_str)
            .replace("{action_desc}", act_desc_str)
            .replace("{action_params}", params_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def precondition_feedback(
        self,
        model: BaseLLM,
        domain_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        parameter: OrderedDict = None,
        preconditions: str = None,
        action_name: str = None,
        action_desc: str = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
        functions: list[Function] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of a PDDL action precondition.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            parameter (OrderedDict): PDDL params of current action
            preconditions (str): PDDL precondition of current action
            action_name (str): name of action
            action_desc (str): description of action
            types (dict[str,str] | list[dict[str,str]]): dictionary of types currently in specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of predicates currently in specification
            functions (list[Function]): list of functions currently in specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        act_name_str = action_name if action_name else "No action name provided"
        act_desc_str = action_desc if action_desc else "No action description provided"
        params_str = (
            "\n".join([f"{name} - {type}" for name, type in parameter.items()])
            if parameter
            else "No parameters provided"
        )
        prec_str = preconditions if preconditions else "No preconditions provided."

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            feedback_template.replace("{domain_desc}", domain_desc)
            .replace("{action_name}", act_name_str)
            .replace("{action_desc}", act_desc_str)
            .replace("{action_params}", params_str)
            .replace("{action_preconditions}", prec_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def effect_feedback(
        self,
        model: BaseLLM,
        domain_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        parameter: OrderedDict = None,
        preconditions: str = None,
        effects: str = None,
        action_name: str = None,
        action_desc: str = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
        functions: list[Function] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of a PDDL action effect.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            parameter (OrderedDict): PDDL params of current action
            preconditions (str): PDDL precondition of current action
            effects (str): PDDL effect of current action
            action_name (str): name of action
            action_desc (str): description of action
            types (dict[str,str] | list[dict[str,str]]): dictionary of types currently in specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of predicates currently in specification
            functions (list[Function]): list of functions currently in specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        act_name_str = action_name if action_name else "No action name provided"
        act_desc_str = action_desc if action_desc else "No action description provided"
        params_str = (
            "\n".join([f"{name} - {type}" for name, type in parameter.items()])
            if parameter
            else "No parameters provided"
        )
        prec_str = preconditions if preconditions else "No preconditions provided."
        eff_str = effects if effects else "No effects provided."

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            feedback_template.replace("{domain_desc}", domain_desc)
            .replace("{action_name}", act_name_str)
            .replace("{action_desc}", act_desc_str)
            .replace("{action_params}", params_str)
            .replace("{action_preconditions}", prec_str)
            .replace("{action_effects}", eff_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def predicate_feedback(
        self,
        model: BaseLLM,
        domain_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of PDDL predicates.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            types (dict[str,str] | list[dict[str,str]]): dictionary of types currently in specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of predicates currently in specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )

        prompt = (
            feedback_template.replace("{domain_desc}", domain_desc)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def task_feedback(
        self,
        model: BaseLLM,
        problem_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
        functions: list[Function] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of a PDDL task.

        Args:
            model (BaseLLM): LLM to query
            problem_desc (str): general problem information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            objects (dict[str,str]): objects of current task specification
            initial (list[dict[str,str]]): initial states of current task specification
            goal (list[dict[str,str]]): goal states of current task specification
            types (dict[str,str] | list[dict[str,str]]): dictionary of types currently in specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of predicates currently in specification
            functions (list[Function]): list of functions currently in specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        obj_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )
        init_str = format_initial(initial) if initial else "No initial state provided."
        goal_str = format_goal(goal) if goal else "No goal state provided."

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            feedback_template.replace("{problem_desc}", problem_desc)
            .replace("{objects}", obj_str)
            .replace("{initial_states}", init_str)
            .replace("{goal_states}", goal_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def objects_feedback(
        self,
        model: BaseLLM,
        problem_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        objects: dict[str, str] = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
        functions: list[Function] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of PDDL task objects.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            objects (dict[str,str]): objects of current task specification
            types (dict[str,str] | list[dict[str,str]]): dictionary of types currently in specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of predicates currently in specification
            functions (list[Function]): list of functions currently in specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        obj_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            feedback_template.replace("{problem_desc}", problem_desc)
            .replace("{objects}", obj_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def initial_state_feedback(
        self,
        model: BaseLLM,
        problem_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
        functions: list[Function] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of PDDL task initial states.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            objects (dict[str,str]): objects of current task specification
            initial (list[dict[str,str]]): initial states of current task specification
            types (dict[str,str] | list[dict[str,str]]): dictionary of types currently in specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of predicates currently in specification
            functions (list[Function]): list of functions currently in specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        obj_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )
        init_str = format_initial(initial) if initial else "No initial state provided."

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            feedback_template.replace("{problem_desc}", problem_desc)
            .replace("{objects}", obj_str)
            .replace("{initial_states}", init_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg

    @require_llm
    def goal_state_feedback(
        self,
        model: BaseLLM,
        problem_desc: str,
        feedback_template: str,
        feedback_type: str = "llm",
        llm_output: str = "",
        objects: dict[str, str] = None,
        initial: list[dict[str, str]] = None,
        goal: list[dict[str, str]] = None,
        types: dict[str, str] | list[dict[str, str]] = None,
        constants: dict[str, str] = None,
        predicates: list[Predicate] = None,
        functions: list[Function] = None,
    ) -> tuple[bool, str]:
        """
        Provides feedback to initial LLM output of PDDL task goal states.

        Args:
            model (BaseLLM): LLM to query
            domain_desc (str): general domain information
            feedback_template (str): prompt template to guide LLM to provide feedback to initial output
            feedback_type (str): type of feedback assistant - 'llm', 'human'
            llm_output (str): original LLM output
            objects (dict[str,str]): objects of current task specification
            initial (list[dict[str,str]]): initial states of current task specification
            goal (list[dict[str,str]]): goal states of current task specification
            types (dict[str,str] | list[dict[str,str]]): dictionary of types currently in specification
            constants (dict[str,str]): current constants in specification, defaults to None
            predicates (list[Predicate]): list of predicates currently in specification
            functions (list[Function]): list of functions currently in specification

        Returns:
            no_fb (bool): flag that deems if feedback is not needed
            fb_msg (str): feedback message from assistant
        """

        # format string info replacements
        obj_str = (
            "\n".join([f"{obj} - {type}" for obj, type in objects.items()])
            if objects
            else "No objects provided."
        )
        init_str = format_initial(initial) if initial else "No initial state provided."
        goal_str = format_goal(goal) if goal else "No goal state provided."

        types_str = pretty_print_dict(types) if types else "No types provided."
        const_str = (
            format_constants(constants) if constants else "No constants provided."
        )
        preds_str = (
            "\n".join([f"{pred['raw']}" for pred in predicates])
            if predicates
            else "No predicates provided."
        )
        funcs_str = (
            "\n".join([f"{func['raw']}" for func in functions])
            if functions
            else "No functions provided."
        )

        prompt = (
            feedback_template.replace("{problem_desc}", problem_desc)
            .replace("{objects}", obj_str)
            .replace("{initial_states}", init_str)
            .replace("{goal_states}", goal_str)
            .replace("{types}", types_str)
            .replace("{constants}", const_str)
            .replace("{predicates}", preds_str)
            .replace("{functions}", funcs_str)
            .replace("{llm_output}", llm_output)
        )

        no_fb, fb_msg = self.get_feedback(model, prompt, feedback_type, llm_output)

        return no_fb, fb_msg



"""
This file contains code for calling LLMs and saving raw model outputs.
Currently, this builder class contains the generic method to run a wide range of LLMs.
"""

import logging, functools
from abc import ABC, abstractmethod
from typing import Any
import yaml

LOG: logging.Logger = logging.getLogger(__name__)


def load_yaml(config_path: str) -> dict[str, Any]:
    with open(config_path, "r") as f:
        return yaml.safe_load(f)


def require_llm(func):
    """
    Decorator to check if an LLM instance is provided and catch errors.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # check if LLM instance is provided
        model = kwargs.get("model", None)
        if model is None:
            # attempt tp get model from positional argument
            for arg in args:
                if isinstance(arg, BaseLLM):
                    model = arg
                    break

        if not model:
            raise ValueError("An LLM instance must be provided to use this method.")

        # try-catch to ensure LLM is even used properly
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(
                f"An error occurred in {func.__name__}: {e}.\n You must provide an LLM engine OR proper configuration to use L2P. Refer to https://github.com/AI-Planning/l2p."
            )
            raise

    return wrapper


class BaseLLM(ABC):
    def __init__(self, model: str, api_key: str | None = None) -> None:

        if not self.valid_models():
            raise ValueError(
                f"No valid models exist in '{self.provider}' provider. Please check if your provider is spelt correctly, or exists in your llm.yaml configuration."
            )

        if model not in self.valid_models():
            raise ValueError(
                f"'{model}' is not a valid model for '{self.provider}' provider. "
                f"Valid models are: {', '.join(self.valid_models())}."
            )
        self.model: str = model
        self.api_key: str | None = api_key

    @abstractmethod
    def query(self, prompt: str) -> str:
        """
        Abstract method to query an LLM with a given prompt and return the response.

        Args:
            prompt (str): The prompt to send to the LLM
        Returns:
            str: The response from the LLM
        """
        pass

    def query_with_system_prompt(self, system_prompt: str, prompt: str) -> str:
        """
        Abstract method to query an LLM with a given prompt and system prompt and return the response.

        Args:
            system_prompt (str): The system prompt to send to the LLM
            prompt (str): The prompt to send to the LLM
        Returns:
            str: The response from the LLM
        """
        return self.query(system_prompt + "\n" + prompt)

    def valid_models(self) -> list[str]:
        """
        List of valid model parameters, e.g., 'gpt4o-mini' for GPT
        """
        return []



"""
This is a subclass (OPENAI) for abstract class (BaseLLM) that implements an interface
to interact with OpenAI's chat-based language models via the official OpenAI API. It
supports providers that are compatible with OpenAI SDK usage (i.e. DeepSeek, Anthropic, etc.).

A YAML configuration file is required to specify model parameters, costs, and other
provider-specific settings. By default, the l2p library includes a configuration file
located at 'l2p/llm/utils/llm.yaml'.

Users can also define their own custom models and parameters by extending the YAML
configuration using the same format template.
"""

from retry import retry
from typing_extensions import override
from .base import BaseLLM, load_yaml


class OPENAI(BaseLLM):
    def __init__(
        self,
        model: str,
        config_path: str = "l2p/llm/utils/llm.yaml",
        provider: str = "openai",
        api_key: str | None = None,
        base_url: str = "https://api.openai.com/v1/",
    ) -> None:

        # load yaml configuration path
        self.provider = provider
        self._config = load_yaml(config_path)

        # attempt to import necessary OPENAI modules
        try:
            from openai import OpenAI
        except ImportError:
            raise ImportError(
                "The 'openai' library is required for OPENAI but is not installed. "
                "Install it using: `pip install openai`."
            )

        try:
            import tiktoken
        except ImportError:
            raise ImportError(
                "The 'tiktoken' library is required for token processing but is not installed. "
                "Install it using: `pip install tiktoken`."
            )

        # retrieve model configurations
        model_config = self._config.get(self.provider, {}).get(model, {})
        self.model_engine = model_config.get("engine", model)

        # call the parent class constructor to handle model and api_key
        super().__init__(model, api_key)
        self.client = OpenAI(api_key=api_key, base_url=base_url)

        # set model parameters
        self._set_parameters(model_config)

        # initialize tokenizer and metadata storage
        self.tok = tiktoken.get_encoding("cl100k_base")
        self.in_tokens = 0
        self.out_tokens = 0
        self.query_log = []  # per-query metadata storage

        # Retrieve cost information for the model from the YAML
        self.cost_per_input_token = model_config.get("cost_usd_mtok", {}).get(
            "input", 0
        )
        self.cost_per_output_token = model_config.get("cost_usd_mtok", {}).get(
            "output", 0
        )

    def _set_parameters(self, model_config: dict) -> None:
        """Set parameters from the model configuration."""

        # default values for parameters if none exists
        defaults = {
            "context_length": 4096,
            "max_completion_tokens": 4096,
            "temperature": 0.0,
            "top_p": 1.0,
            "frequency_penalty": 0.0,
            "presence_penalty": 0.0,
            "stop": None,
            "reasoning_effort": None,
        }

        parameters = model_config.get("model_params", {})
        for key, default in defaults.items():
            setattr(self, key, parameters.get(key, default))

    @retry(tries=2, delay=60)
    def connect_openai(self, client, model, messages, **kwargs):
        """Send a request to OpenAI API"""

        return client.chat.completions.create(model=model, messages=messages, **kwargs)

    @override
    def query(
        self,
        prompt: str,
        messages=None,
        end_when_error=False,
        max_retry=3,
        est_margin=200,
    ) -> str:
        """Generate a response from OpenAI based on the prompt."""

        if not isinstance(prompt, str) or not prompt.strip():
            raise ValueError("Prompt must be a non-empty string.")

        messages = messages or [{"role": "user", "content": prompt}]

        # estimate current usage of tokens
        current_tokens = sum(len(self.tok.encode(m["content"])) for m in messages)
        requested_tokens = min(
            self.max_completion_tokens,
            self.context_length - current_tokens - est_margin,
        )

        print(
            f"Requesting {requested_tokens} tokens "
            f"(estimated prompt: {current_tokens} tokens, margin: {est_margin}, window: {self.context_length})"
        )

        # request response
        conn_success, n_retry = False, 0
        while not conn_success and n_retry < max_retry:
            try:
                print(
                    f"[INFO] connecting to {self.model_engine} ({requested_tokens} tokens)..."
                )

                kwargs = {
                    "temperature": self.temperature,
                    "max_completion_tokens": requested_tokens,
                    "top_p": self.top_p,
                    "frequency_penalty": self.frequency_penalty,
                    "presence_penalty": self.presence_penalty,
                    "stop": self.stop,
                }

                # only add reasoning_effort if it is not None
                if self.reasoning_effort is not None:
                    kwargs["reasoning_effort"] = self.reasoning_effort

                # retrieve completion
                response = self.connect_openai(
                    client=self.client,
                    model=self.model_engine,
                    messages=messages,
                    **kwargs,
                )

                llm_output = response.choices[
                    0
                ].message.content  # retrieve output from completion

                # record token usage
                usage = getattr(response, "usage", None)
                if usage:
                    self.in_tokens += usage.prompt_tokens
                    self.out_tokens += usage.completion_tokens
                else:
                    self.in_tokens += current_tokens
                    self.out_tokens += len(self.tok.encode(llm_output))

                # calculate cost (USD) per million tokens
                input_cost = (self.in_tokens / 1_000_000) * self.cost_per_input_token
                output_cost = (self.out_tokens / 1_000_000) * self.cost_per_output_token
                total_cost = input_cost + output_cost

                self.reset_tokens() # reset tokens after each query

                conn_success = True

            except Exception as e:
                print(f"[ERROR] LLM error: {e}")
                if end_when_error:
                    break

            n_retry += 1

        if not conn_success:
            raise ConnectionError(
                f"Failed to connect to the LLM after {max_retry} retries"
            )

        # log query information
        self.query_log.append(
            {
                "model": self.model_engine,
                "prompt_tokens": usage.prompt_tokens if usage else current_tokens,
                "completion_tokens": (
                    usage.completion_tokens
                    if usage
                    else len(self.tok.encode(llm_output))
                ),
                "reasoning_tokens": (
                    usage.completion_tokens_details.reasoning_tokens if usage else 0
                ),
                "total_tokens": (
                    usage.total_tokens
                    if usage
                    else current_tokens + len(self.tok.encode(llm_output))
                ),
                "input_cost_usd": input_cost,
                "output_cost_usd": output_cost,
                "total_cost_usd": total_cost,
                "messages": messages,
                "output": llm_output,
            }
        )

        return llm_output

    def get_tokens(self) -> tuple[int, int]:
        """Return input and output token counts."""
        return self.in_tokens, self.out_tokens

    def reset_tokens(self) -> None:
        """Reset token counts."""
        self.in_tokens = 0
        self.out_tokens = 0

    def get_query_log(self) -> list:
        """Retrieve query log."""
        return self.query_log

    def reset_query_log(self) -> None:
        """Reset query log."""
        self.query_log = []

    @override
    def valid_models(self) -> list[str]:
        """Returns a list of valid model engines."""
        try:
            return list(self._config.get(self.provider, {}).keys())
        except KeyError:
            return []
